r"""Implement an analyzer that generates a section to analyze the
temporal distribution of null values for all columns."""

from __future__ import annotations

__all__ = ["TemporalNullValueAnalyzer"]

import logging
from typing import TYPE_CHECKING

from coola.utils import repr_indent, repr_mapping

from flamme.analyzer.base import BaseAnalyzer
from flamme.section import EmptySection, TemporalNullValueSection

if TYPE_CHECKING:
    from collections.abc import Sequence

    import pandas as pd


logger = logging.getLogger(__name__)


class TemporalNullValueAnalyzer(BaseAnalyzer):
    r"""Implement an analyzer to show the temporal distribution of null
    values for all columns.

    Args:
        dt_column: The datetime column used to analyze
            the temporal distribution.
        period: The temporal period e.g. monthly or
            daily.
        figsize: The figure size in inches. The first
            dimension is the width and the second is the height.

    Example usage:

    ```pycon
    >>> import numpy as np
    >>> import pandas as pd
    >>> from flamme.analyzer import TemporalNullValueAnalyzer
    >>> analyzer = TemporalNullValueAnalyzer(dt_column="datetime", period="M")
    >>> analyzer
    TemporalNullValueAnalyzer(
      (columns): None
      (dt_column): datetime
      (period): M
      (figsize): None
    )
    >>> frame = pd.DataFrame(
    ...     {
    ...         "col": np.array([np.nan, 1, 0, 1]),
    ...         "datetime": pd.to_datetime(
    ...             ["2020-01-03", "2020-02-03", "2020-03-03", "2020-04-03"]
    ...         ),
    ...     }
    ... )
    >>> section = analyzer.analyze(frame)

    ```
    """

    def __init__(
        self,
        dt_column: str,
        period: str,
        columns: Sequence[str] | None = None,
        figsize: tuple[float, float] | None = None,
    ) -> None:
        self._dt_column = dt_column
        self._period = period
        self._columns = columns
        self._figsize = figsize

    def __repr__(self) -> str:
        args = repr_indent(
            repr_mapping(
                {
                    "columns": self._columns,
                    "dt_column": self._dt_column,
                    "period": self._period,
                    "figsize": self._figsize,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def analyze(self, frame: pd.DataFrame) -> TemporalNullValueSection | EmptySection:
        logger.info(
            f"Analyzing the temporal null value distribution | "
            f"datetime column: {self._dt_column} | period: {self._period}"
        )
        if self._dt_column not in frame:
            logger.info(
                "Skipping temporal null value analysis because the datetime column "
                f"({self._dt_column}) is not in the DataFrame: {sorted(frame.columns)}"
            )
            return EmptySection()
        columns = self._columns
        if columns is None:
            # Exclude the datetime column because it does not make sense to analyze it because
            # we cannot know the date/time if the value is null.
            columns = sorted([col for col in frame.columns if col != self._dt_column])
        return TemporalNullValueSection(
            frame=frame,
            columns=columns,
            dt_column=self._dt_column,
            period=self._period,
            figsize=self._figsize,
        )



r"""Contain the implementation of a CSV ingestor."""

from __future__ import annotations

__all__ = ["CsvIngestor"]

import logging
from typing import TYPE_CHECKING, Any

import pandas as pd

from flamme.ingestor.base import BaseIngestor
from flamme.utils.path import human_file_size, sanitize_path

if TYPE_CHECKING:
    from pathlib import Path

logger = logging.getLogger(__name__)


class CsvIngestor(BaseIngestor):
    r"""Implement a CSV DataFrame ingestor.

    Args:
        path: The path to the CSV file to ingest.
        **kwargs: Additional keyword arguments for
            ``pandas.read_csv``.

    Example usage:

    ```pycon
    >>> from flamme.ingestor import ParquetIngestor
    >>> ingestor = ParquetIngestor(path="/path/to/frame.csv")
    >>> ingestor
    ParquetIngestor(path=/path/to/frame.csv)
    >>> frame = ingestor.ingest()  # doctest: +SKIP

    ```
    """

    def __init__(self, path: Path | str, **kwargs: Any) -> None:
        self._path = sanitize_path(path)
        self._kwargs = kwargs

    def __repr__(self) -> str:
        args = ", ".join([f"{key}={value}" for key, value in self._kwargs.items()])
        if args:
            args = ", " + args
        return f"{self.__class__.__qualname__}(path={self._path}{args})"

    def ingest(self) -> pd.DataFrame:
        logger.info(f"Ingesting CSV data from {self._path} (size={human_file_size(self._path)})...")
        frame = pd.read_csv(filepath_or_buffer=self._path, **self._kwargs)
        logger.info(f"Data ingested. DataFrame shape: {frame.shape}")
        return frame



r"""Contain utility functions to manage matplotlib figures."""

from __future__ import annotations

__all__ = ["figure2html"]

import base64
import io
import math
from typing import TYPE_CHECKING

from matplotlib import pyplot as plt

if TYPE_CHECKING:
    from matplotlib.axes import Axes


def figure2html(fig: plt.Figure, reactive: bool = True, close_fig: bool = False) -> str:
    r"""Convert a matplotlib figure to a string that can be used in a
    HTML file.

    Args:
        fig: The figure to convert.
        reactive: If ``True``, the generated is configured to be
            reactive to the screen size.
        close_fig: If ``True``, the figure is closed after it is
            converted to HTML format.

    Returns:
        The converted figure to a string.

    Example usage:

    ```pycon
    >>> from matplotlib import pyplot as plt
    >>> from flamme.utils.figure import figure2html
    >>> fig, ax = plt.subplots()
    >>> string = figure2html(fig)

    ```
    """
    fig.tight_layout()
    img = io.BytesIO()
    fig.savefig(img, format="png", bbox_inches="tight")
    img.seek(0)
    data = base64.b64encode(img.getvalue()).decode("utf-8")
    if close_fig:
        plt.close(fig)
    style = 'style="width:100%; height:auto;" ' if reactive else False
    return f'<img {style}src="data:image/png;charset=utf-8;base64, {data}">'


def readable_xticklabels(
    ax: Axes,
    max_num_xticks: int = 100,
    xticklabel_max_len: int = 20,
    xticklabel_min: int = 10,
) -> None:
    r"""Update the tick labels to make them easier to read, in particular
    if the tick labels are dense.

    Args:
        ax: The figure axes to update.
        max_num_xticks: The maximum number of ticks to show in the
            figure.
        xticklabel_max_len: If a tick label has a length greater than
            this value, the tick labels are rotated vertically.
        xticklabel_min: If the number of ticks is lower than this
            number the tick labels are rotated vertically.

    Example usage:

    ```pycon
    >>> import numpy as np
    >>> from matplotlib import pyplot as plt
    >>> from flamme.utils.figure import readable_xticklabels
    >>> fig, ax = plt.subplots()
    >>> ax.hist(np.arange(10), bins=10)
    >>> readable_xticklabels(ax)

    ```
    """
    xticks = ax.get_xticks()
    if len(xticks) > max_num_xticks:
        n = math.ceil(len(xticks) / max_num_xticks)
        xticks = xticks[::n]
        ax.set_xticks(xticks)
    if len(xticks) > xticklabel_min or any(
        len(str(label)) > xticklabel_max_len for label in ax.get_xticklabels()
    ):
        ax.tick_params(axis="x", labelrotation=90)



r"""Contain ``pandas.Series`` transformers to transform columns with
string values."""

from __future__ import annotations

__all__ = ["StripStringSeriesTransformer"]

from typing import TYPE_CHECKING

from flamme.transformer.series.base import BaseSeriesTransformer

if TYPE_CHECKING:
    import pandas as pd


class StripStringSeriesTransformer(BaseSeriesTransformer):
    r"""Implement a transformer to strip the strings in a
    ``pandas.Series``.

    Example usage:

    ```pycon
    >>> import pandas as pd
    >>> from flamme.transformer.series import StripString
    >>> transformer = StripString()
    >>> transformer
    StripStringSeriesTransformer()
    >>> series = pd.Series(["a ", " b", "  c  ", " d ", "e"])
    >>> out = transformer.transform(series)
    >>> out
    0    a
    1    b
    2    c
    3    d
    4    e
    dtype: object

    ```
    """

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def transform(self, series: pd.Series) -> pd.Series:
        return series.apply(lambda x: x.strip() if isinstance(x, str) else x)



r"""Contain the implementation of a section to analyze the data types of
each column."""

from __future__ import annotations

__all__ = ["DataTypeSection"]

import copy
import logging
from typing import TYPE_CHECKING

from coola.utils import repr_indent, repr_mapping
from jinja2 import Template

from flamme.section.base import BaseSection
from flamme.section.utils import (
    GO_TO_TOP,
    render_html_toc,
    tags2id,
    tags2title,
    valid_h_tag,
)

if TYPE_CHECKING:
    from collections.abc import Sequence

    from numpy.typing import DTypeLike


logger = logging.getLogger(__name__)


class DataTypeSection(BaseSection):
    r"""Implement a section that analyzes the data type of each column.

    Args:
        dtypes: The data type for each column.
        types: The types of the values in each
            column. A column can contain multiple types. The keys are
            the column names.

    Example usage:

    ```pycon
    >>> import numpy as np
    >>> from flamme.section import DataTypeSection
    >>> section = DataTypeSection(
    ...     dtypes={
    ...         "float": np.dtype("float64"),
    ...         "int": np.dtype("float64"),
    ...         "str": np.dtype("O"),
    ...     },
    ...     types={"float": {float}, "int": {int}, "str": {str, type(None)}},
    ... )
    >>> section
    DataTypeSection(
      (dtypes): {'float': dtype('float64'), 'int': dtype('float64'), 'str': dtype('O')}
      (types): {'float': {<class 'float'>}, 'int': {<class 'int'>}, 'str': {<class 'NoneType'>, <class 'str'>}}
    )
    >>> section.get_statistics()
    {'float': {<class 'float'>}, 'int': {<class 'int'>}, 'str': {<class 'NoneType'>, <class 'str'>}}

    ```
    """

    def __init__(self, dtypes: dict[str, DTypeLike], types: dict[str, set]) -> None:
        self._dtypes = dtypes
        self._types = types

        dtkeys = set(self._dtypes.keys())
        tkeys = set(self._types.keys())
        if dtkeys != tkeys:
            msg = (
                f"The keys of dtypes and types do not match:\n"
                f"({len(dtkeys)}): {dtkeys}\n"
                f"({len(tkeys)}): {tkeys}\n"
            )
            raise RuntimeError(msg)

    def __repr__(self) -> str:
        args = repr_indent(repr_mapping({"dtypes": self._dtypes, "types": self._types}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def get_statistics(self) -> dict:
        return copy.deepcopy(self._types)

    def render_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        return Template(self._create_template()).render(
            {
                "go_to_top": GO_TO_TOP,
                "id": tags2id(tags),
                "depth": valid_h_tag(depth + 1),
                "title": tags2title(tags),
                "section": number,
                "table": self._create_table(),
            }
        )

    def render_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        logger.info("Rendering the data types report...")
        return render_html_toc(number=number, tags=tags, depth=depth, max_depth=max_depth)

    def _create_template(self) -> str:
        return """
<h{{depth}} id="{{id}}">{{section}} {{title}} </h{{depth}}>

{{go_to_top}}

<p style="margin-top: 1rem;">
This section analyzes the values types for each column.

<ul>
  <li> <b>data type</b>: is the pandas data type used to represent the column </li>
  <li> <b>types</b>: are the real object types for the objects in the column </li>
</ul>

{{table}}
"""

    def _create_table(self) -> str:
        columns = sorted(self._types.keys())
        rows = "\n".join(
            [
                create_table_row(column=col, types=self._types[col], dtype=self._dtypes[col])
                for col in columns
            ]
        )
        return Template(
            """
<table class="table table-hover table-responsive w-auto" >
    <thead class="thead table-group-divider">
        <tr>
            <th>column</th>
            <th>data type</th>
            <th>types</th>
        </tr>
    </thead>
    <tbody class="tbody table-group-divider">
        {{rows}}
        <tr class="table-group-divider"></tr>
    </tbody>
</table>
"""
        ).render({"rows": rows})


def create_table_row(column: str, dtype: DTypeLike, types: set) -> str:
    r"""Create the HTML code of a new table row.

    Args:
        column: The column name.
        dtype: The column data type.
        types: The types in th column.

    Returns:
        The HTML code of a row.
    """
    types = sorted([str(t).replace("<", "&lt;").replace(">", "&gt;") for t in types])
    return Template(
        """<tr>
    <th>{{column}}</th>
    <td>{{dtype}}</td>
    <td>{{types}}</td>
</tr>"""
    ).render({"column": column, "dtype": dtype, "types": ", ".join(types)})



r"""Contain the implementation of a section to generate a summary of a
DataFrame."""

from __future__ import annotations

__all__ = ["DataFrameSummarySection"]

import logging
from typing import TYPE_CHECKING, Any

from jinja2 import Template

from flamme.section.base import BaseSection
from flamme.section.utils import (
    GO_TO_TOP,
    render_html_toc,
    tags2id,
    tags2title,
    valid_h_tag,
)
from flamme.utils.dtype import series_column_types

if TYPE_CHECKING:
    from collections.abc import Sequence

    import pandas as pd


logger = logging.getLogger(__name__)


class DataFrameSummarySection(BaseSection):
    r"""Implement a section that returns a summary of a DataFrame.

    Args:
        frame: The DataFrame to analyze.
        top: The number of most frequent values to show.

    Example usage:

    ```pycon
    >>> import pandas as pd
    >>> import numpy as np
    >>> from flamme.section import DataFrameSummarySection
    >>> section = DataFrameSummarySection(
    ...     frame=pd.DataFrame(
    ...         {
    ...             "col1": np.array([1.2, 4.2, 4.2, 2.2]),
    ...             "col2": np.array([1, 1, 1, 1]),
    ...             "col3": np.array([1, 2, 2, 2]),
    ...         }
    ...     )
    ... )
    >>> section
    DataFrameSummarySection(top=5)
    >>> section.get_statistics()
    {'columns': ('col1', 'col2', 'col3'), 'null_count': (0, 0, 0), 'nunique': (3, 1, 2), 'column_types': ({<class 'float'>}, {<class 'int'>}, {<class 'int'>})}
    """

    def __init__(self, frame: pd.DataFrame, top: int = 5) -> None:
        self._frame = frame
        if top < 0:
            msg = f"Incorrect top value ({top}). top must be positive"
            raise ValueError(msg)
        self._top = top

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(top={self._top})"

    @property
    def frame(self) -> pd.DataFrame:
        r"""The DataFrame to analyze."""
        return self._frame

    @property
    def top(self) -> int:
        return self._top

    def get_columns(self) -> tuple[str, ...]:
        return tuple(self._frame.columns)

    def get_null_count(self) -> tuple[int, ...]:
        return tuple(
            self._frame.isna().sum().to_frame("__count__")["__count__"].astype(int).tolist()
        )

    def get_nunique(self) -> tuple[int, ...]:
        return tuple(self._frame.nunique(dropna=False).astype(int).tolist())

    def get_column_types(self) -> tuple[set[type], ...]:
        return tuple(series_column_types(self._frame[col]) for col in self.frame)

    def get_most_frequent_values(self, top: int = 5) -> tuple[tuple[tuple[Any, int], ...], ...]:
        value_counts = []
        for col in self.frame:
            values = self._frame[col].value_counts(dropna=False, sort=True).head(top)
            value_counts.append(tuple((val, c) for val, c in zip(values.index, values.to_list())))
        return tuple(value_counts)

    def get_statistics(self) -> dict:
        return {
            "columns": self.get_columns(),
            "null_count": self.get_null_count(),
            "nunique": self.get_nunique(),
            "column_types": self.get_column_types(),
        }

    def render_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        logger.info("Rendering the DataFrame summary section...")
        return Template(self._create_template()).render(
            {
                "go_to_top": GO_TO_TOP,
                "id": tags2id(tags),
                "depth": valid_h_tag(depth + 1),
                "title": tags2title(tags),
                "section": number,
                "table": self._create_table(),
                "nrows": f"{self._frame.shape[0]:,}",
                "ncols": f"{self._frame.shape[1]:,}",
            }
        )

    def render_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        return render_html_toc(number=number, tags=tags, depth=depth, max_depth=max_depth)

    def _create_template(self) -> str:
        return """
<h{{depth}} id="{{id}}">{{section}} {{title}} </h{{depth}}>

{{go_to_top}}

<p style="margin-top: 1rem;">
This section shows a short summary of each column.

<ul>
  <li> <b>column</b>: are the column names</li>
  <li> <b>types</b>: are the real object types for the objects in the column </li>
  <li> <b>null</b>: are the number (and percentage) of null values in the column </li>
  <li> <b>unique</b>: are the number (and percentage) of unique values in the column </li>
</ul>

<p style="margin-top: 1rem;">
<b>General statistics about the DataFrame</b>

<ul>
  <li> number of rows: {{nrows}}</li>
  <li> number of columns: {{ncols}} </li>
</ul>

{{table}}
<p style="margin-top: 1rem;">
"""

    def _create_table(self) -> str:
        total = self._frame.shape[0]
        return create_table(
            columns=self.get_columns(),
            null_count=self.get_null_count(),
            nunique=self.get_nunique(),
            column_types=self.get_column_types(),
            most_frequent_values=self.get_most_frequent_values(top=self._top),
            total=total,
        )


def create_table(
    columns: Sequence[str],
    null_count: Sequence[int],
    nunique: Sequence[int],
    column_types: Sequence[set],
    most_frequent_values: Sequence[Sequence[tuple[Any, int]]],
    total: int,
) -> str:
    rows = []
    for (
        column,
        null,
        nuniq,
        types,
        mf_values,
    ) in zip(columns, null_count, nunique, column_types, most_frequent_values):
        rows.append(
            create_table_row(
                column=column,
                null=null,
                types=types,
                nunique=nuniq,
                most_frequent_values=mf_values,
                total=total,
            )
        )
    rows = "\n".join(rows)
    return Template(
        """
<table class="table table-hover table-responsive w-auto" >
<thead class="thead table-group-divider">
    <tr>
        <th>column</th>
        <th>types</th>
        <th>null</th>
        <th>unique</th>
        <th>most frequent values</th>
    </tr>
</thead>
<tbody class="tbody table-group-divider">
    {{rows}}
    <tr class="table-group-divider"></tr>
</tbody>
</table>
"""
    ).render({"rows": rows})


def create_table_row(
    column: str,
    null: int,
    nunique: int,
    types: set,
    most_frequent_values: Sequence[tuple[Any, int]],
    total: int,
) -> str:
    r"""Create the HTML code of a new table row.

    Args:
        column: The column name.
        null: The number of null values.
        nunique: The number of unique values.
        types: The types in th column.
        most_frequent_values: The most frequent values.
        total: The total number of rows.

    Returns:
        The HTML code of a row.
    """
    total = max(total, 1)
    types = ", ".join(sorted([prepare_type_name(t) for t in types]))
    null = f"{null:,} ({100 * null / total:.2f}%)"
    nunique = f"{nunique:,} ({100 * nunique / total:.2f}%)"
    most_frequent_values = ", ".join(
        [f"{val} ({100 * c / total:.2f}%)" for val, c in most_frequent_values]
    )
    return Template(
        """<tr>
    <th>{{column}}</th>
    <td>{{types}}</td>
    <td {{num_style}}>{{null}}</td>
    <td {{num_style}}>{{nunique}}</td>
    <td>{{most_frequent_values}}</td>
</tr>"""
    ).render(
        {
            "num_style": 'style="text-align: right;"',
            "column": column,
            "null": null,
            "types": types,
            "nunique": nunique,
            "most_frequent_values": most_frequent_values,
        }
    )


TYPE_NAMES = {
    "pandas._libs.tslibs.timestamps.Timestamp": "pandas.Timestamp",
    "pandas._libs.tslibs.nattype.NaTType": "pandas.NaTType",
    "pandas._libs.missing.NAType": "pandas.NAType",
}


def prepare_type_name(typ: type) -> str:
    r"""Return a compact type name when possible.

    Args:
        typ: The input type.

    Returns:
        The compact type name.
    """
    name = str(typ).split("'", maxsplit=2)[1].rsplit("'", maxsplit=2)[0]
    # Reduce the name if possible
    return TYPE_NAMES.get(name, name)



r"""Contain the implementation of an empty section."""

from __future__ import annotations

__all__ = ["EmptySection"]

from typing import TYPE_CHECKING

from flamme.section.base import BaseSection

if TYPE_CHECKING:
    from collections.abc import Sequence


class EmptySection(BaseSection):
    r"""Implement an empty section.

    This section is implemented to deal with missing columns or to skip
    some analyses.

    Example usage:

    ```pycon
    >>> import pandas as pd
    >>> import numpy as np
    >>> from flamme.section import EmptySection
    >>> section = EmptySection()
    >>> section
    EmptySection()
    >>> section.get_statistics()
    {}

    ```
    """

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def get_statistics(self) -> dict:
        return {}

    def render_html_body(
        self,
        number: str = "",  # noqa: ARG002
        tags: Sequence[str] = (),  # noqa: ARG002
        depth: int = 0,  # noqa: ARG002
    ) -> str:
        return ""

    def render_html_toc(
        self,
        number: str = "",  # noqa: ARG002
        tags: Sequence[str] = (),  # noqa: ARG002
        depth: int = 0,  # noqa: ARG002
        max_depth: int = 1,  # noqa: ARG002
    ) -> str:
        return ""



r"""Contain the implementation of a clickhouse ingestor."""

from __future__ import annotations

__all__ = ["ClickHouseIngestor"]

import logging
from typing import TYPE_CHECKING

from flamme.ingestor.base import BaseIngestor
from flamme.utils import setup_object

if TYPE_CHECKING:
    import pandas as pd

    from flamme.utils.imports import is_clickhouse_connect_available

    if is_clickhouse_connect_available():
        import clickhouse_connect


logger = logging.getLogger(__name__)


class ClickHouseIngestor(BaseIngestor):
    r"""Implement a clickhouse DataFrame ingestor.

    Args:
        query: The query to get the data.
        client: The clickhouse client or its configuration.
            Please check the documentation of
            ``clickhouse_connect.get_client`` to get more information.

    Example usage:

    ```pycon
    >>> from flamme.ingestor import ClickHouseIngestor
    >>> import clickhouse_connect
    >>> client = clickhouse_connect.get_client()  # doctest: +SKIP
    >>> ingestor = ClickHouseIngestor(query="", client=client)  # doctest: +SKIP
    >>> frame = ingestor.ingest()  # doctest: +SKIP

    ```
    """

    def __init__(self, query: str, client: clickhouse_connect.driver.Client | dict) -> None:
        self._query = str(query)
        self._client: clickhouse_connect.driver.Client = setup_object(client)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def ingest(self) -> pd.DataFrame:
        logger.info(
            f"Ingesting data from clickhouse... \n\n"
            "---------------------------------------------------------------------------------\n"
            f"query:\n{self._query}\n"
            "---------------------------------------------------------------------------------\n\n"
        )
        frame = self._client.query_df(query=self._query).sort_index(axis=1)
        logger.info(f"Data ingested. DataFrame shape: {frame.shape}")
        logger.info(f"number of unique column names: {len(set(frame.columns)):,}")
        return frame



r"""Contain the base class to implement an reader."""

from __future__ import annotations

__all__ = ["BaseSchemaReader", "is_schema_reader_config", "setup_schema_reader"]

import logging
from abc import ABC
from typing import TYPE_CHECKING

from objectory import AbstractFactory
from objectory.utils import is_object_config

if TYPE_CHECKING:
    import pyarrow as pa

logger = logging.getLogger(__name__)


class BaseSchemaReader(ABC, metaclass=AbstractFactory):
    r"""Define the base class to implement a schema reader.

    Example usage:

    ```pycon

    >>> import tempfile
    >>> import pandas as pd
    >>> from pathlib import Path
    >>> from flamme.schema.reader import ParquetSchemaReader
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     path = Path(tmpdir).joinpath("data.parquet")
    ...     pd.DataFrame({"col1": [1, 2, 3], "col2": ["a", "b", "c"]}).to_parquet(
    ...         path, index=False
    ...     )
    ...     reader = ParquetSchemaReader(path)
    ...     reader
    ...     schema = reader.read()
    ...     schema
    ...
    ParquetSchemaReader(path=.../data.parquet)
    col1: int64
    col2: string
    ...

    ```
    """

    def read(self) -> pa.Schema:
        r"""Read the schema associated to a DataFrame.

        Returns:
            The ingested DataFrame.

        Example usage:

        ```pycon

        >>> import tempfile
        >>> import pandas as pd
        >>> from pathlib import Path
        >>> from flamme.schema.reader import ParquetSchemaReader
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     path = Path(tmpdir).joinpath("data.parquet")
        ...     pd.DataFrame({"col1": [1, 2, 3], "col2": ["a", "b", "c"]}).to_parquet(path, index=False)
        ...     reader = ParquetSchemaReader(path)
        ...     schema = reader.read()
        ...     schema
        ...
        col1: int64
        col2: string
        ...

        ```
        """


def is_schema_reader_config(config: dict) -> bool:
    r"""Indicate if the input configuration is a configuration for a
    ``BaseSchemaReader``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
        config: The configuration to check.

    Returns:
        ``True`` if the input configuration is a configuration
            for a ``BaseSchemaReader`` object.

    Example usage:

    ```pycon
    >>> from flamme.schema.reader import is_schema_reader_config
    >>> is_schema_reader_config(
    ...     {"_target_": "flamme.schema.reader.ParquetSchemaReader", "path": "/path/to/data.parquet"}
    ... )
    True

    ```
    """
    return is_object_config(config, BaseSchemaReader)


def setup_schema_reader(
    reader: BaseSchemaReader | dict,
) -> BaseSchemaReader:
    r"""Set up a schema reader.

    The reader is instantiated from its configuration
    by using the ``BaseSchemaReader`` factory function.

    Args:
        reader: Specifies a schema reader or its configuration.

    Returns:
        An instantiated schema reader.

    Example usage:

    ```pycon
    >>> from flamme.schema.reader import setup_schema_reader
    >>> reader = setup_schema_reader(
    ...     {"_target_": "flamme.schema.reader.ParquetSchemaReader", "path": "/path/to/data.parquet"}
    ... )
    >>> reader
    ParquetSchemaReader(path=.../data.parquet)

    ```
    """
    if isinstance(reader, dict):
        logger.info("Initializing a schema reader from its configuration... ")
        reader = BaseSchemaReader.factory(**reader)
    if not isinstance(reader, BaseSchemaReader):
        logger.warning(f"reader is not a `BaseSchemaReader` (received: {type(reader)})")
    return reader



r"""Contain the implementation of a section to manage a dictionary of
sections."""

from __future__ import annotations

__all__ = ["SectionDict"]

from typing import TYPE_CHECKING

from coola.utils import repr_indent, repr_mapping, str_indent

from flamme.section.base import BaseSection
from flamme.section.utils import (
    GO_TO_TOP,
    render_html_toc,
    tags2id,
    tags2title,
    valid_h_tag,
)

if TYPE_CHECKING:
    from collections.abc import Sequence


class SectionDict(BaseSection):
    r"""Implement a section to manage a dictionary of sections.

    Args:
        sections: The dictionary of sections.
        max_toc_depth: The maximum level to show in the
            table of content. Set this value to ``0`` to not show
            the table of content at the beginning of the section.

    Example usage:

    ```pycon
    >>> import pandas as pd
    >>> from flamme.section import SectionDict, ContentSection, TemporalRowCountSection
    >>> frame = pd.DataFrame(
    ...     {
    ...         "datetime": pd.to_datetime(
    ...             [
    ...                 "2020-01-03",
    ...                 "2020-01-04",
    ...                 "2020-01-05",
    ...                 "2020-02-03",
    ...                 "2020-03-03",
    ...                 "2020-04-03",
    ...             ]
    ...         )
    ...     }
    ... )
    >>> section = SectionDict(
    ...     {
    ...         "content": ContentSection("meow"),
    ...         "rows": TemporalRowCountSection(frame, dt_column="datetime", period="M"),
    ...     }
    ... )
    >>> section
    SectionDict(
      (content): ContentSection()
      (rows): TemporalRowCountSection(dt_column=datetime, period=M, figsize=None)
    )
    >>> section.get_statistics()
    {'content': {}, 'rows': {}}

    ```
    """

    def __init__(self, sections: dict[str, BaseSection], max_toc_depth: int = 0) -> None:
        self._sections = sections
        self._max_toc_depth = max_toc_depth

    def __repr__(self) -> str:
        args = repr_indent(repr_mapping(self._sections))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    @property
    def sections(self) -> dict[str, BaseSection]:
        return self._sections

    @property
    def max_toc_depth(self) -> int:
        return self._max_toc_depth

    def get_statistics(self) -> dict:
        return {name: section.get_statistics() for name, section in self._sections.items()}

    def render_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        report = []
        if tags:
            report.append(
                f'<h{valid_h_tag(depth + 1)} id="{tags2id(tags)}">{number} '
                f"{tags2title(tags)} </h{valid_h_tag(depth + 1)}>"
            )
            report.extend([GO_TO_TOP, '<p style="margin-top: 1rem;">'])

        if self._max_toc_depth > 0:
            report.append(
                self._render_html_toc_subsections(
                    number=number, tags=tags, depth=0, max_depth=self._max_toc_depth
                )
            )

        for i, (name, section) in enumerate(self._sections.items()):
            report.append(
                section.render_html_body(
                    number=f"{number}{i + 1}.", tags=[*list(tags), name], depth=depth + 1
                )
            )
        return "\n".join(report)

    def render_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        if depth >= max_depth:
            return ""
        toc = []
        if tags:
            toc.append(render_html_toc(number=number, tags=tags, depth=depth, max_depth=max_depth))
        subtoc = self._render_html_toc_subsections(tags=tags, depth=depth + 1, max_depth=max_depth)
        if subtoc:
            toc.append(subtoc)
        return "\n".join(toc)

    def _render_html_toc_subsections(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        subtoc = []
        for i, (name, section) in enumerate(self._sections.items()):
            line = section.render_html_toc(
                number=f"{number}{i + 1}.",
                tags=[*list(tags), name],
                depth=depth,
                max_depth=max_depth,
            )
            if line:
                subtoc.append(f"  {str_indent(line)}")
        if subtoc:
            subtoc.insert(0, "<ul>")
            subtoc.append("</ul>")
        return "\n".join(subtoc)



r"""Contain utility functions to filter columns in DataFrames."""

from __future__ import annotations

__all__ = [
    "find_columns_decimal",
    "find_columns_str",
    "find_columns_type",
]

from decimal import Decimal
from typing import TYPE_CHECKING

from flamme.utils.dtype import frame_column_types

if TYPE_CHECKING:
    import pandas as pd


def find_columns_type(frame: pd.DataFrame, cls: type) -> tuple[str, ...]:
    r"""Find the list of columns that contains a given type.

    Args:
        frame: The DataFrame to filter.
        cls: The type to find.

    Returns:
        tuple: The tuple of columns with the given type.

    Example usage:

    ```pycon
    >>> import pandas as pd
    >>> from flamme.utils.filtering import find_columns_type
    >>> frame = pd.DataFrame(
    ...     {
    ...         "col1": [1, 2, 3, 4, 5],
    ...         "col2": ["1", "2", "3", "4", "5"],
    ...         "col3": ["1", "2", "3", "4", "5"],
    ...         "col4": ["a", "b", "c", "d", "e"],
    ...     }
    ... )
    >>> find_columns_type(frame, str)
    ('col2', 'col3', 'col4')

    ```
    """
    types = frame_column_types(frame)
    return tuple(col for col, tps in types.items() if cls in tps)


def find_columns_decimal(frame: pd.DataFrame) -> tuple[str, ...]:
    r"""Find the list of columns that contains the type string.

    Args:
        frame: The DataFrame.

    Returns:
        The tuple of columns with the type string.

    Example usage:

    ```pycon
    >>> import pandas as pd
    >>> from decimal import Decimal
    >>> from flamme.utils.filtering import find_columns_decimal
    >>> frame = pd.DataFrame(
    ...     {
    ...         "col1": [1, 2, 3, 4, Decimal(5)],
    ...         "col2": ["1", "2", "3", "4", "5"],
    ...         "col3": ["1", "2", "3", "4", "5"],
    ...         "col4": ["a", "b", "c", "d", "e"],
    ...     }
    ... )
    >>> find_columns_decimal(frame)
    ('col1',)

    ```
    """
    return find_columns_type(frame, Decimal)


def find_columns_str(frame: pd.DataFrame) -> tuple[str, ...]:
    r"""Find the list of columns that contains the type string.

    Args:
        frame: The input DataFrame.

    Returns:
        The tuple of columns with the type string.

    Example usage:

    ```pycon
    >>> import numpy as np
    >>> import pandas as pd
    >>> from flamme.utils.filtering import find_columns_str
    >>> frame = pd.DataFrame(
    ...     {
    ...         "col1": [1, 2, 3, 4, 5],
    ...         "col2": ["1", "2", "3", "4", "5"],
    ...         "col3": ["1", "2", "3", "4", "5"],
    ...         "col4": ["a", "b", "c", "d", "e"],
    ...     }
    ... )
    >>> find_columns_str(frame)
    ('col2', 'col3', 'col4')

    ```
    """
    return find_columns_type(frame, str)



r"""Implement an analyzer to analyze only a subset of the columns."""

from __future__ import annotations

__all__ = ["ContentAnalyzer"]

import logging
from typing import TYPE_CHECKING

from flamme.analyzer.base import BaseAnalyzer
from flamme.section import ContentSection

if TYPE_CHECKING:

    import pandas as pd

    from flamme.section import BaseSection

logger = logging.getLogger(__name__)


class ContentAnalyzer(BaseAnalyzer):
    r"""Implement an analyzer that generates the given custom content.

    Args:
        content: The content to use in the HTML code.

    Example usage:

    ```pycon
    >>> import numpy as np
    >>> import pandas as pd
    >>> from flamme.analyzer import ContentAnalyzer
    >>> analyzer = ContentAnalyzer(content="meow")
    >>> analyzer
    ContentAnalyzer()
    >>> frame = pd.DataFrame(
    ...     {
    ...         "int": np.array([np.nan, 1, 0, 1]),
    ...         "float": np.array([1.2, 4.2, np.nan, 2.2]),
    ...         "str": np.array(["A", "B", None, np.nan]),
    ...     }
    ... )
    >>> section = analyzer.analyze(frame)

    ```
    """

    def __init__(self, content: str) -> None:
        self._content = str(content)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def analyze(self, frame: pd.DataFrame) -> BaseSection:  # noqa: ARG002
        logger.info("Generating the given custom content...")
        return ContentSection(content=self._content)



r"""Contain the implementation of a section to analyze a column with
discrete values."""

from __future__ import annotations

__all__ = ["ColumnDiscreteSection"]

import logging
from typing import TYPE_CHECKING

import numpy as np
from coola.utils import repr_indent, repr_mapping
from jinja2 import Template
from matplotlib import pyplot as plt

from flamme.section.base import BaseSection
from flamme.section.utils import (
    GO_TO_TOP,
    render_html_toc,
    tags2id,
    tags2title,
    valid_h_tag,
)
from flamme.utils.figure import figure2html, readable_xticklabels

if TYPE_CHECKING:
    from collections import Counter
    from collections.abc import Sequence

logger = logging.getLogger(__name__)


class ColumnDiscreteSection(BaseSection):
    r"""Implement a section that analyzes a discrete distribution of
    values.

    Args:
        counter: The counter that represents the discrete
            distribution.
        null_values: The number of null values.
        column: The column name.
        max_rows: The maximum number of rows to show in the
            table.
        yscale: The y-axis scale. If ``'auto'``, the
            ``'linear'`` or ``'log'`` scale is chosen based on the
            distribution.
        figsize: The figure size in inches. The first
            dimension is the width and the second is the height.

    Example usage:

    ```pycon
    >>> from collections import Counter
    >>> from flamme.section import ColumnDiscreteSection
    >>> section = ColumnDiscreteSection(counter=Counter({"a": 4, "b": 2, "c": 6}), column="col")
    >>> section
    ColumnDiscreteSection(
      (null_values): 0
      (column): col
      (yscale): auto
      (max_rows): 20
      (figsize): None
    )
    >>> section.get_statistics()
    {'most_common': [('c', 6), ('a', 4), ('b', 2)], 'nunique': 3, 'total': 12}

    ```
    """

    def __init__(
        self,
        counter: Counter,
        null_values: int = 0,
        column: str = "N/A",
        max_rows: int = 20,
        yscale: str = "auto",
        figsize: tuple[float, float] | None = None,
    ) -> None:
        self._counter = counter
        self._null_values = null_values
        self._column = column
        self._max_rows = int(max_rows)
        self._yscale = yscale
        self._figsize = figsize

        self._total = sum(self._counter.values())

    def __repr__(self) -> str:
        args = repr_indent(
            repr_mapping(
                {
                    "null_values": self._null_values,
                    "column": self._column,
                    "yscale": self._yscale,
                    "max_rows": self._max_rows,
                    "figsize": self._figsize,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    @property
    def figsize(self) -> tuple[float, float] | None:
        r"""The individual figure size in pixels.

        The first dimension is the width and the second is the height.
        """
        return self._figsize

    @property
    def yscale(self) -> str:
        return self._yscale

    def get_statistics(self) -> dict:
        most_common = [(value, count) for value, count in self._counter.most_common() if count > 0]
        return {
            "most_common": most_common,
            "nunique": len(most_common),
            "total": self._total,
        }

    def render_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        logger.info(f"Rendering the discrete distribution of {self._column}")
        stats = self.get_statistics()
        null_values_pct = (
            f"{100 * self._null_values / stats['total']:.2f}" if stats["total"] > 0 else "N/A"
        )
        return Template(self._create_template()).render(
            {
                "go_to_top": GO_TO_TOP,
                "id": tags2id(tags),
                "depth": valid_h_tag(depth + 1),
                "title": tags2title(tags),
                "section": number,
                "column": self._column,
                "total_values": f"{stats['total']:,}",
                "unique_values": f"{stats['nunique']:,}",
                "null_values": f"{self._null_values:,}",
                "null_values_pct": null_values_pct,
                "figure": self._create_figure(),
                "table": self._create_table(),
            }
        )

    def render_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        return render_html_toc(number=number, tags=tags, depth=depth, max_depth=max_depth)

    def _create_template(self) -> str:
        return """
<h{{depth}} id="{{id}}">{{section}} {{title}} </h{{depth}}>

{{go_to_top}}

<p style="margin-top: 1rem;">
This section analyzes the discrete distribution of values for column <em>{{column}}</em>.

<ul>
  <li> total values: {{total_values}} </li>
  <li> number of unique values: {{unique_values}} </li>
  <li> number of null values: {{null_values}} / {{total_values}} ({{null_values_pct}}%) </li>
</ul>

{{figure}}
{{table}}
<p style="margin-top: 1rem;">
"""

    def _create_figure(self) -> str:
        if self._total == 0:
            return "<span>&#9888;</span> No figure is generated because the column is empty"
        most_common = [(value, count) for value, count in self._counter.most_common() if count > 0]
        fig = create_histogram(
            column=self._column,
            labels=[str(value) for value, _ in most_common],
            counts=[count for _, count in most_common],
            yscale=self._yscale,
            figsize=self._figsize,
        )
        return Template(
            r"""
<p style="margin-top: 1rem;">
<b>Distribution of values in column {{column}}</b>

<p>The values in the figure below are sorted by decreasing order of number of occurrences.

{{figure}}
"""
        ).render({"figure": fig, "column": self._column})

    def _create_table(self) -> str:
        if self._total == 0:
            return "<span>&#9888;</span> No table is generated because the column is empty"

        most_common = self._counter.most_common(self._max_rows)
        rows_head = "\n".join(
            [create_table_row(column=col, count=count) for col, count in most_common]
        )
        lest_common = self._counter.most_common()[-self._max_rows :][::-1]
        rows_tail = "\n".join(
            [create_table_row(column=col, count=count) for col, count in lest_common]
        )
        return Template(
            """
<details>
    <summary>[show head and tail values]</summary>

    <div class="row">
      <div class="col">
        <p style="margin-top: 1rem;">
        <b>Head: {{max_values}} most common values in column <em>{{column}}</em></b>
        <table class="table table-hover table-responsive w-auto" >
            <thead class="thead table-group-divider">
                <tr>
                    <th>column</th>
                    <th>count</th>
                </tr>
            </thead>
            <tbody class="tbody table-group-divider">
                {{rows_head}}
                <tr class="table-group-divider"></tr>
            </tbody>
        </table>
      </div>
      <div class="col">
        <p style="margin-top: 1rem;">
        <b>Tail: {{max_values}} least common values in column <em>{{column}}</em></b>
        <table class="table table-hover table-responsive w-auto" >
            <thead class="thead table-group-divider">
                <tr>
                    <th>column</th>
                    <th>count</th>
                </tr>
            </thead>
            <tbody class="tbody table-group-divider">
                {{rows_tail}}
                <tr class="table-group-divider"></tr>
            </tbody>
        </table>
      </div>
    </div>
</details>
"""
        ).render(
            {
                "max_values": len(most_common),
                "rows_head": rows_head,
                "rows_tail": rows_tail,
                "column": self._column,
            }
        )


def create_histogram(
    column: str,
    labels: list[str],
    counts: list[int],
    yscale: str = "auto",
    figsize: tuple[float, float] | None = None,
) -> str:
    fig, ax = plt.subplots(figsize=figsize)
    x = np.arange(len(labels))
    ax.bar(x, counts, width=0.9 if len(labels) < 50 else 1, color="tab:blue")
    if yscale == "auto":
        yscale = "log" if (max(counts) / min(counts)) >= 50 else "linear"
    ax.set_yscale(yscale)
    ax.set_xticks(x, labels=labels)
    readable_xticklabels(ax, max_num_xticks=100)
    ax.set_xlim(-0.5, len(labels) - 0.5)
    ax.set_ylabel("Number of occurrences")
    ax.set_title(f"Number of occurrences for each value of {column}")
    return figure2html(fig, close_fig=True)


def create_table_row(column: str, count: int) -> str:
    r"""Create the HTML code of a new table row.

    Args:
        column: The column name.
        count (int): The count for the column.

    Returns:
        The HTML code of a row.
    """
    return Template("""<tr><th>{{column}}</th><td {{num_style}}>{{count}}</td></tr>""").render(
        {"num_style": 'style="text-align: right;"', "column": column, "count": f"{count:,}"}
    )



r"""Contain utility functions to manage ranges of values."""

from __future__ import annotations

__all__ = ["find_range"]

import numpy as np


def find_range(
    values: np.ndarray,
    xmin: float | str | None = None,
    xmax: float | str | None = None,
) -> tuple[float, float]:
    r"""Find a valid range of value.

    Args:
        values: The values used to find the quantiles.
        xmin: The minimum value of the range or its
            associated quantile. ``q0.1`` means the 10% quantile.
            ``0`` is the minimum value and ``1`` is the maximum value.
        xmax: The maximum value of the range or its
            associated quantile. ``q0.9`` means the 90% quantile.
            ``0`` is the minimum value and ``1`` is the maximum value.

    Returns:
        The range of values in the format ``(min, max)``.

    Example usage:

    ```pycon
    >>> import numpy as np
    >>> from flamme.utils.range import find_range
    >>> data = np.arange(101)
    >>> find_range(data)
    (0, 100)
    >>> find_range(data, xmin=5, xmax=50)
    (5, 50)
    >>> find_range(data, xmin="q0.1", xmax="q0.9")
    (10.0, 90.0)

    ```
    """
    if xmin is None:
        xmin = np.nanmin(values).item()
    if xmax is None:
        xmax = np.nanmax(values).item()
    q = [float(x[1:]) for x in [xmin, xmax] if isinstance(x, str)]
    quantiles = np.nanquantile(values, q)
    if isinstance(xmin, str):
        xmin = quantiles[0]
    if isinstance(xmax, str):
        xmax = quantiles[-1]
    if isinstance(xmin, np.number):
        xmin = xmin.item()
    if isinstance(xmax, np.number):
        xmax = xmax.item()
    return (xmin, xmax)



r"""Implement an analyzer that generates a table of content section."""

from __future__ import annotations

__all__ = ["TableOfContentAnalyzer"]

import logging
from typing import TYPE_CHECKING

from coola.utils import str_indent, str_mapping

from flamme.analyzer.base import BaseAnalyzer, setup_analyzer
from flamme.section.toc import TableOfContentSection

if TYPE_CHECKING:
    import pandas as pd

logger = logging.getLogger(__name__)


class TableOfContentAnalyzer(BaseAnalyzer):
    r"""Implement a wrapper around an analyzer to add a table of content
    to the generated section report.

    Args:
        analyzer: The analyzer or its configuration.
        max_toc_depth: The maximum level to show in the
            table of content.

    Example usage:

    ```pycon
    >>> import numpy as np
    >>> import pandas as pd
    >>> from flamme.analyzer import TableOfContentAnalyzer, DuplicatedRowAnalyzer
    >>> analyzer = TableOfContentAnalyzer(DuplicatedRowAnalyzer())
    >>> analyzer
    TableOfContentAnalyzer(
      (analyzer): DuplicatedRowAnalyzer(columns=None, figsize=None)
      (max_toc_depth): 1
    )
    >>> frame = pd.DataFrame(
    ...     {
    ...         "col1": np.array([0, 1, 0, 1]),
    ...         "col2": np.array([1, 0, 1, 0]),
    ...         "col3": np.array([1, 1, 1, 1]),
    ...     }
    ... )
    >>> section = analyzer.analyze(frame)

    ```
    """

    def __init__(
        self,
        analyzer: BaseAnalyzer | dict,
        max_toc_depth: int = 1,
    ) -> None:
        self._analyzer = setup_analyzer(analyzer)
        self._max_toc_depth = max_toc_depth

    def __repr__(self) -> str:
        args = str_indent(
            str_mapping({"analyzer": self._analyzer, "max_toc_depth": self._max_toc_depth})
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def analyze(self, frame: pd.DataFrame) -> TableOfContentSection:
        return TableOfContentSection(
            section=self._analyzer.analyze(frame), max_toc_depth=self._max_toc_depth
        )



r"""Implement an analyzer to automatically selects an analyzer based on
a given selection logic."""

from __future__ import annotations

__all__ = ["ChoiceAnalyzer", "NumUniqueSelection"]

from collections.abc import Callable, Mapping
from typing import TYPE_CHECKING

from coola.utils import str_indent, str_mapping

from flamme.analyzer.base import BaseAnalyzer, setup_analyzer

if TYPE_CHECKING:

    import pandas as pd

    from flamme.section import BaseSection


class ChoiceAnalyzer(BaseAnalyzer):
    r"""Implement an analyzer to analyze multiple analyzers.

    Args:
        analyzers: The mappings to analyze. The key of each analyzer
            is used to organize the metrics and report.
        selection_fn: Specifies a callable with the selection logic.
            The callable returns the key of the analyzer to use based
            on the data in the input DataFrame.

    Example usage:

    ```pycon
    >>> import numpy as np
    >>> import pandas as pd
    >>> from flamme.analyzer import (
    ...     ChoiceAnalyzer,
    ...     FilteredAnalyzer,
    ...     NullValueAnalyzer,
    ...     DuplicatedRowAnalyzer,
    ... )
    >>> analyzer = ChoiceAnalyzer(
    ...     {"null": NullValueAnalyzer(), "duplicate": DuplicatedRowAnalyzer()},
    ...     selection_fn=lambda frame: "null" if frame.isna().values.any() else "duplicate",
    ... )
    >>> analyzer
    ChoiceAnalyzer(
      (null): NullValueAnalyzer(figsize=None)
      (duplicate): DuplicatedRowAnalyzer(columns=None, figsize=None)
    )
    >>> frame = pd.DataFrame(
    ...     {
    ...         "int": np.array([np.nan, 1, 0, 1]),
    ...         "float": np.array([1.2, 4.2, np.nan, 2.2]),
    ...         "str": np.array(["A", "B", None, np.nan]),
    ...     }
    ... )
    >>> section = analyzer.analyze(frame)
    >>> section.__class__.__qualname__
    NullValueSection
    >>> frame = pd.DataFrame({"col": np.arange(10)})
    >>> section = analyzer.analyze(frame)
    >>> section.__class__.__qualname__
    DuplicatedRowSection

    ```
    """

    def __init__(
        self,
        analyzers: Mapping[str, BaseAnalyzer | dict],
        selection_fn: Callable[[pd.DataFrame], str],
    ) -> None:
        self._analyzers = {name: setup_analyzer(analyzer) for name, analyzer in analyzers.items()}
        self._selection_fn = selection_fn

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(\n  {str_indent(str_mapping(self._analyzers))}\n)"

    @property
    def analyzers(self) -> dict[str, BaseAnalyzer]:
        return self._analyzers

    def analyze(self, frame: pd.DataFrame) -> BaseSection:
        analyzer = self._analyzers[self._selection_fn(frame)]
        return analyzer.analyze(frame)


class NumUniqueSelection(Callable):
    r"""Implement a selection logic based on the number of unique values
    in a column.

    Args:
        column: The column to check.
        threshold: The threshold of number of unique values.
        small: The string to return if the number of unique
            values is lower than the threshold.
        large: The string to return if the number of unique
            values is greater than the threshold.

    Example usage:

    ```pycon
    >>> import numpy as np
    >>> import pandas as pd
    >>> from flamme.analyzer.choice import NumUniqueSelection
    >>> selection = NumUniqueSelection(column='col')
    >>> selection
    NumUniqueSelection(column=col, threshold=100, small=small, large=large)
    >>> selection(pd.DataFrame({'col': np.arange(10)}))
    small
    >>> selection(pd.DataFrame({'col': np.arange(1000)}))
    large
    """

    def __init__(
        self, column: str, threshold: int = 100, small: str = "small", large: str = "large"
    ) -> None:
        self._column = column
        self._threshold = threshold
        self._small = small
        self._large = large

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(column={self._column}, threshold={self._threshold}, "
            f"small={self._small}, large={self._large})"
        )

    def __call__(self, frame: pd.DataFrame) -> str:
        nunique = frame[self._column].nunique()
        return self._small if nunique <= self._threshold else self._large



r"""Contain the implementation of a simple ingestor."""

from __future__ import annotations

__all__ = ["ClickHouseSchemaReader"]

import logging
from typing import TYPE_CHECKING

from flamme.schema.reader.base import BaseSchemaReader
from flamme.utils import setup_object

if TYPE_CHECKING:
    import pyarrow as pa

    from flamme.utils.imports import is_clickhouse_connect_available

    if is_clickhouse_connect_available():
        import clickhouse_connect

logger = logging.getLogger(__name__)


class ClickHouseSchemaReader(BaseSchemaReader):
    r"""Implement a simple DataFrame ingestor.

    Args:
        query: The query to get the data.
        client: The clickhouse client or its configuration.
            Please check the documentation of
            ``clickhouse_connect.get_client`` to get more information.

    Example usage:

    ```pycon
    >>> import pandas as pd
    >>> from flamme.schema.reader import ClickHouseSchemaReader
    >>> client = clickhouse_connect.get_client()  # doctest: +SKIP
    >>> reader = ClickHouseSchemaReader(query="", client=client)  # doctest: +SKIP
    >>> schema = reader.read()  # doctest: +SKIP

    ```
    """

    def __init__(self, query: str, client: clickhouse_connect.driver.Client | dict) -> None:
        self._query = str(query)
        self._client: clickhouse_connect.driver.Client = setup_object(client)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def read(self) -> pa.Schema:
        logger.info(
            f"reading schema from clickhouse using the following query... \n\n"
            "---------------------------------------------------------------------------------\n"
            f"{self._query}\n"
            "---------------------------------------------------------------------------------\n\n"
        )
        return self._client.query_arrow(query=self._query).schema



r"""Contain the implementation of a simple ingestor."""

from __future__ import annotations

__all__ = ["SchemaReader"]

from typing import TYPE_CHECKING

import pyarrow as pa

from flamme.schema.reader.base import BaseSchemaReader

if TYPE_CHECKING:
    import pandas as pd


class SchemaReader(BaseSchemaReader):
    r"""Implement a simple DataFrame ingestor.

    Args:
        frame: The DataFrame to ingest.

    Example usage:

    ```pycon
    >>> import pandas as pd
    >>> from flamme.schema.reader import SchemaReader
    >>> reader = SchemaReader(
    ...     frame=pd.DataFrame(
    ...         {
    ...             "col1": [1, 2, 3, 4, 5],
    ...             "col2": [1.1, 2.2, 3.3, 4.4, 5.5],
    ...             "col4": ["a", "b", "c", "d", "e"],
    ...         }
    ...     )
    ... )
    >>> reader
    SchemaReader(shape=(5, 3))
    >>> schema = reader.read()
    >>> schema
    col1: int64
    col2: double
    col4: string
    ...

    ```
    """

    def __init__(self, frame: pd.DataFrame) -> None:
        self._frame = frame

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(shape={self._frame.shape})"

    def read(self) -> pa.Schema:
        return pa.Schema.from_pandas(self._frame)



r"""Contain a ``pandas.DataFrame`` transformer to combine sequentially
multiple transformers."""

from __future__ import annotations

__all__ = ["SequentialDataFrameTransformer"]

from typing import TYPE_CHECKING

from coola.utils import str_indent, str_sequence

from flamme.transformer.dataframe.base import (
    BaseDataFrameTransformer,
    setup_dataframe_transformer,
)

if TYPE_CHECKING:
    from collections.abc import Sequence

    import pandas as pd


class SequentialDataFrameTransformer(BaseDataFrameTransformer):
    r"""Implement a ``pandas.DataFrame`` transformer to apply
    sequentially several transformers.

    Args:
        transformers: The transformers or their
            configurations.

    Example usage:

    ```pycon
    >>> import pandas as pd
    >>> from flamme.transformer.dataframe import (
    ...     Sequential,
    ...     StripString,
    ...     ToNumeric,
    ... )
    >>> transformer = Sequential(
    ...     [
    ...         StripString(columns=["col1", "col3"]),
    ...         ToNumeric(columns=["col1", "col2"]),
    ...     ]
    ... )
    >>> transformer
    SequentialDataFrameTransformer(
      (0): StripStringDataFrameTransformer(columns=('col1', 'col3'))
      (1): ToNumericDataFrameTransformer(columns=('col1', 'col2'))
    )
    >>> frame = pd.DataFrame(
    ...     {
    ...         "col1": [1, 2, 3, 4, 5],
    ...         "col2": ["1", "2", "3", "4", "5"],
    ...         "col3": ["a ", " b", "  c  ", "d", "e"],
    ...         "col4": ["a ", " b", "  c  ", "d", "e"],
    ...     }
    ... )
    >>> frame
       col1 col2   col3   col4
    0     1    1     a      a
    1     2    2      b      b
    2     3    3    c      c
    3     4    4      d      d
    4     5    5      e      e
    >>> out = transformer.transform(frame)
    >>> out
       col1 col2 col3   col4
    0     1    1    a     a
    1     2    2    b      b
    2     3    3    c    c
    3     4    4    d      d
    4     5    5    e      e

    ```
    """

    def __init__(self, transformers: Sequence[BaseDataFrameTransformer | dict]) -> None:
        self._transformers = tuple(
            setup_dataframe_transformer(transformer) for transformer in transformers
        )

    def __repr__(self) -> str:
        args = ""
        if self._transformers:
            args = f"\n  {str_indent(str_sequence(self._transformers))}\n"
        return f"{self.__class__.__qualname__}({args})"

    def transform(self, frame: pd.DataFrame) -> pd.DataFrame:
        for transformer in self._transformers:
            frame = transformer.transform(frame)
        return frame



r"""Implement an analyzer that generates a markdown section."""

from __future__ import annotations

__all__ = ["MarkdownAnalyzer"]

from typing import TYPE_CHECKING

from flamme.analyzer.base import BaseAnalyzer
from flamme.section import MarkdownSection

if TYPE_CHECKING:
    import pandas as pd


class MarkdownAnalyzer(BaseAnalyzer):
    r"""Implement an analyzer that adds a mardown string to the report.

    Args:
        desc: The markdown description.

    Example usage:

    ```pycon
    >>> import numpy as np
    >>> import pandas as pd
    >>> from flamme.analyzer import MarkdownAnalyzer
    >>> analyzer = MarkdownAnalyzer(desc="hello cats!")
    >>> analyzer
    MarkdownAnalyzer()
    >>> frame = pd.DataFrame({})
    >>> section = analyzer.analyze(frame)

    ```
    """

    def __init__(self, desc: str) -> None:
        self._desc = str(desc)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def analyze(self, frame: pd.DataFrame) -> MarkdownSection:  # noqa: ARG002
        return MarkdownSection(desc=self._desc)



r"""Contain IO functions to load/save data."""

from __future__ import annotations

__all__ = ["load_text", "save_text"]

import logging
from pathlib import Path

logger = logging.getLogger(__name__)


def load_text(path: Path) -> str:
    r"""Read the data from a given text file.

    Args:
        path: The path where to the text file.

    Returns:
        The data from the text file.

    Example usage:

    ```pycon
    >>> from pathlib import Path
    >>> from flamme.utils.io import load_text
    >>> data = load_text(Path("/path/to/data.txt"))  # xdoctest: +SKIP()

    ```
    """
    logger.debug(f"read {path}")
    with Path.open(path) as file:
        return file.read()


def save_text(to_save: str, path: Path) -> None:
    r"""Save the given data in a text file.

    Args:
        to_save: The data to write in a text file.
        path: The path where to write the text file.

    Example usage:

    ```pycon
    >>> from pathlib import Path
    >>> from flamme.utils.io import save_text
    >>> save_text("abc", Path("/path/to/data.txt"))  # xdoctest: +SKIP()

    ```
    """
    logger.debug(f"write data in a text file: {path}")
    path.parents[0].mkdir(exist_ok=True, parents=True)
    # Save to tmp, then commit by moving the file in case the job gets
    # interrupted while writing the file
    tmp_path = path.parents[0].joinpath(f"{path.name}.tmp")
    with Path.open(tmp_path, mode="w") as file:
        file.write(to_save)
    tmp_path.rename(path)



r"""Contain the implementation of a section to analyze the most frequent
values for a given columns."""

from __future__ import annotations

__all__ = ["MostFrequentValuesSection"]

import logging
from typing import TYPE_CHECKING

from coola.utils import repr_indent, repr_mapping
from jinja2 import Template

from flamme.section.base import BaseSection
from flamme.section.utils import (
    GO_TO_TOP,
    render_html_toc,
    tags2id,
    tags2title,
    valid_h_tag,
)

if TYPE_CHECKING:
    from collections import Counter
    from collections.abc import Sequence

logger = logging.getLogger(__name__)


class MostFrequentValuesSection(BaseSection):
    r"""Implement a section that analyzes the most frequent values for a
    given columns.

    Args:
        counter: The counter with the number of occurrences
            for all values.
        column: The column name.
        top: The maximum number of values to show.

    Example usage:

    ```pycon
    >>> from collections import Counter
    >>> from flamme.section import MostFrequentValuesSection
    >>> section = MostFrequentValuesSection(
    ...     counter=Counter({"a": 4, "b": 2, "c": 6}), column="col"
    ... )
    >>> section
    MostFrequentValuesSection(
      (counter): Counter({'c': 6, 'a': 4, 'b': 2})
      (column): col
      (top): 100
      (total): 12
    )
    >>> section.get_statistics()
    {'most_common': [('c', 6), ('a', 4), ('b', 2)]}

    ```
    """

    def __init__(self, counter: Counter, column: str, top: int = 100) -> None:
        self._counter = counter
        self._column = column
        self._top = top

        self._total = sum(self._counter.values())

    def __repr__(self) -> str:
        args = repr_indent(
            repr_mapping(
                {
                    "counter": self._counter,
                    "column": self._column,
                    "top": self._top,
                    "total": self._total,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def get_statistics(self) -> dict:
        return {"most_common": list(self._counter.most_common(self._top))}

    def render_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        logger.info("Rendering the most frequent values section...")
        return Template(self._create_template()).render(
            {
                "go_to_top": GO_TO_TOP,
                "id": tags2id(tags),
                "depth": valid_h_tag(depth + 1),
                "title": tags2title(tags),
                "section": number,
                "table": self._create_table(),
                "column": self._column,
                "top": f"{self._top:,}",
            }
        )

    def render_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        return render_html_toc(number=number, tags=tags, depth=depth, max_depth=max_depth)

    def _create_template(self) -> str:
        return """
<h{{depth}} id="{{id}}">{{section}} {{title}} </h{{depth}}>

{{go_to_top}}

<p style="margin-top: 1rem;">
This section analyzes the {{top}} most frequent values in <em>{{column}}</em>.

<ul>
  <li> <b>count</b>: is the number of occurrences of the value </li>
  <li> <b>percentage</b>: is the number of occurrences the value divided by the total number of occurrences </li>
  <li> <b>cumulative percentage</b>: is the sum of the percentage of occurrence for the value plus all the previous values in the table. </li>
</ul>

{{table}}
<p style="margin-top: 1rem;">
"""

    def _create_table(self) -> str:
        most_common = list(self._counter.most_common(self._top))
        rows = []
        cumcount = 0
        for value, count in most_common:
            cumcount += count
            rows.append(
                create_table_row(value=value, count=count, total=self._total, cumcount=cumcount)
            )
        return Template(
            """
<table class="table table-hover table-responsive w-auto" >
    <thead class="thead table-group-divider">
        <tr>
            <th>value</th>
            <th>count</th>
            <th>percentage (%)</th>
            <th>cumulative percentage (%)</th>
        </tr>
    </thead>
    <tbody class="tbody table-group-divider">
        {{rows}}
        <tr class="table-group-divider"></tr>
    </tbody>
</table>
"""
        ).render({"rows": "\n".join(rows)})


def create_table_row(value: str, count: int, total: int, cumcount: int) -> str:
    r"""Create the HTML code of a new table row.

    Args:
        value: Specifies a string representation of the value.
        count: The number of occurrences of the value.
        total: The total number of occurrences.
        cumcount: The cumulative number of occurrences.

    Returns:
        The HTML code of a row.
    """
    return Template(
        """<tr>
    <th>{{value}}</th>
    <td {{num_style}}>{{count}}</td>
    <td {{num_style}}>{{percentage}}</td>
    <td {{num_style}}>{{cum_percentage}}</td>
</tr>"""
    ).render(
        {
            "num_style": 'style="text-align: right;"',
            "value": value,
            "count": f"{count:,}",
            "percentage": f"{100 * count / total:.2f}",
            "cum_percentage": f"{100 * cumcount / total:.2f}",
        }
    )



r"""Implement an analyzer that generates a section about the number of
rows per temporal window."""

from __future__ import annotations

__all__ = ["TemporalRowCountAnalyzer"]

import logging
from typing import TYPE_CHECKING

from flamme.analyzer.base import BaseAnalyzer
from flamme.section import EmptySection, TemporalRowCountSection

if TYPE_CHECKING:
    import pandas as pd

logger = logging.getLogger(__name__)


class TemporalRowCountAnalyzer(BaseAnalyzer):
    r"""Implement an analyzer to show the number of rows per temporal
    window.

    Args:
        dt_column: The datetime column used to analyze
            the temporal distribution.
        period: The temporal period e.g. monthly or daily.
        figsize: The figure size in inches. The first
            dimension is the width and the second is the height.

    Example usage:

    ```pycon
    >>> import numpy as np
    >>> import pandas as pd
    >>> from flamme.analyzer import TemporalRowCountAnalyzer
    >>> analyzer = TemporalRowCountAnalyzer(dt_column="datetime", period="M")
    >>> analyzer
    TemporalRowCountAnalyzer(dt_column=datetime, period=M, figsize=None)
    >>> frame = pd.DataFrame(
    ...     {
    ...         "datetime": pd.to_datetime(
    ...             ["2020-01-03", "2020-02-03", "2020-03-03", "2020-04-03"]
    ...         ),
    ...     }
    ... )
    >>> section = analyzer.analyze(frame)

    ```
    """

    def __init__(
        self,
        dt_column: str,
        period: str,
        figsize: tuple[float, float] | None = None,
    ) -> None:
        self._dt_column = dt_column
        self._period = period
        self._figsize = figsize

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(dt_column={self._dt_column}, "
            f"period={self._period}, figsize={self._figsize})"
        )

    def analyze(self, frame: pd.DataFrame) -> TemporalRowCountSection | EmptySection:
        logger.info(
            f"Analyzing the number of rows | "
            f"datetime column: {self._dt_column} | period: {self._period}"
        )
        if self._dt_column not in frame:
            logger.info(
                "Skipping number of rows analysis because the datetime column "
                f"({self._dt_column}) is not in the DataFrame: {sorted(frame.columns)}"
            )
            return EmptySection()
        return TemporalRowCountSection(
            frame=frame,
            dt_column=self._dt_column,
            period=self._period,
            figsize=self._figsize,
        )



r"""Contain the base class to implement a ``pandas.Series``
transformer."""

from __future__ import annotations

__all__ = ["BaseSeriesTransformer", "is_series_transformer_config", "setup_series_transformer"]

import logging
from abc import ABC
from typing import TYPE_CHECKING

from objectory import AbstractFactory
from objectory.utils import is_object_config

if TYPE_CHECKING:
    import pandas as pd

logger = logging.getLogger(__name__)


class BaseSeriesTransformer(ABC, metaclass=AbstractFactory):
    r"""Define the base class to transform a ``pandas.Series``.

    Example usage:

    ```pycon
    >>> import pandas as pd
    >>> from flamme.transformer.series import ToNumeric
    >>> transformer = ToNumeric()
    >>> transformer
    ToNumericSeriesTransformer()
    >>> series = pd.Series(["1", "2", "3", "4", "5"])
    >>> series.dtype
    dtype('O')
    >>> out = transformer.transform(series)
    >>> out.dtype
    dtype('int64')

    ```
    """

    def transform(self, series: pd.Series) -> pd.Series:
        r"""Transform a ``pandas.Series``.

        Args:
            series: The ``pandas.Series`` to transform.

        Returns:
            The transformed ``pandas.Series``.

        Example usage:

        ```pycon
        >>> import pandas as pd
        >>> from flamme.transformer.series import ToNumeric
        >>> transformer = ToNumeric()
        >>> series = pd.Series(["1", "2", "3", "4", "5"])
        >>> out = transformer.transform(series)
        >>> out.dtype
        dtype('int64')

        ```
        """


def is_series_transformer_config(config: dict) -> bool:
    r"""Indicate if the input configuration is a configuration for a
    ``BaseSeriesTransformer``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
        config: The configuration to check.

    Returns:
        ``True`` if the input configuration is a configuration
            for a ``BaseSeriesTransformer`` object.

    Example usage:

    ```pycon
    >>> from flamme.transformer.series import is_series_transformer_config
    >>> is_series_transformer_config({"_target_": "flamme.transformer.series.ToNumeric"})
    True

    ```
    """
    return is_object_config(config, BaseSeriesTransformer)


def setup_series_transformer(
    transformer: BaseSeriesTransformer | dict,
) -> BaseSeriesTransformer:
    r"""Set up a ``pandas.Series`` transformer.

    The transformer is instantiated from its configuration
    by using the ``BaseSeriesTransformer`` factory function.

    Args:
        transformer: Specifies a ``pandas.Series`` transformer or its
            configuration.

    Returns:
        An instantiated transformer.

    Example usage:

    ```pycon
    >>> from flamme.transformer.series import setup_series_transformer
    >>> transformer = setup_series_transformer(
    ...     {"_target_": "flamme.transformer.series.ToNumeric"}
    ... )
    >>> transformer
    ToNumericSeriesTransformer()

    ```
    """
    if isinstance(transformer, dict):
        logger.info("Initializing a series transformer from its configuration... ")
        transformer = BaseSeriesTransformer.factory(**transformer)
    if not isinstance(transformer, BaseSeriesTransformer):
        logger.warning(
            f"transformer is not a `BaseSeriesTransformer` (received: {type(transformer)})"
        )
    return transformer



r"""Contain the implementation of a parquet schema reader."""

from __future__ import annotations

__all__ = ["ParquetSchemaReader"]

import logging
from typing import TYPE_CHECKING

from pyarrow.parquet import read_schema

from flamme.schema.reader.base import BaseSchemaReader
from flamme.utils.path import sanitize_path

if TYPE_CHECKING:
    from pathlib import Path

    import pyarrow as pa

logger = logging.getLogger(__name__)


class ParquetSchemaReader(BaseSchemaReader):
    r"""Implement a parquet schema reader.

    Args:
        path: The path to the parquet file to ingest.

    Example usage:

    ```pycon

    >>> import tempfile
    >>> import pandas as pd
    >>> from pathlib import Path
    >>> from flamme.schema.reader import ParquetSchemaReader
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     path = Path(tmpdir).joinpath("data.parquet")
    ...     pd.DataFrame({"col1": [1, 2, 3], "col2": ["a", "b", "c"]}).to_parquet(
    ...         path, index=False
    ...     )
    ...     reader = ParquetSchemaReader(path)
    ...     reader
    ...     schema = reader.read()
    ...     schema
    ...
    ParquetSchemaReader(path=.../data.parquet)
    col1: int64
    col2: string
    ...

    ```
    """

    def __init__(self, path: Path | str) -> None:
        self._path = sanitize_path(path)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(path={self._path})"

    def read(self) -> pa.Schema:
        logger.info(f"reading the schema from parquet file {self._path}...")
        schema = read_schema(self._path)
        logger.info("schema read")
        return schema



r"""Contain utility functions for mappings."""

from __future__ import annotations

__all__ = ["sort_by_keys"]


def sort_by_keys(mapping: dict) -> dict:
    r"""Sorts a dictionary by keys.

    Args:
        mapping: The dictionary to sort.

    Returns:
        dict: The sorted dictionary.

    Example usage:

    ```pycon
    >>> from flamme.utils.mapping import sort_by_keys
    >>> sort_by_keys({"dog": 1, "cat": 5, "fish": 2})
    {'cat': 5, 'dog': 1, 'fish': 2}

    ```
    """
    return dict(sorted(mapping.items()))



r"""Contain very experimental code to manage columns."""

from __future__ import annotations

__all__ = ["BaseColumn", "Column"]

from abc import abstractmethod

from coola.utils import str_indent, str_mapping

from flamme.analyzer.base import BaseAnalyzer, setup_analyzer
from flamme.transformer.series.base import (
    BaseSeriesTransformer,
    setup_series_transformer,
)


class BaseColumn:
    r"""Define the column base class.

    Example usage:

    ```pycon
    >>> from flamme.utils.columns import Column
    >>> from flamme.analyzer import ColumnContinuousAnalyzer
    >>> from flamme.transformer.series import ToNumeric
    >>> column = Column(
    ...     can_be_null=True,
    ...     analyzer=ColumnContinuousAnalyzer(column="col"),
    ...     transformer=ToNumeric(),
    ... )
    >>> column
    Column(
      (can_be_null): True
      (analyzer): ColumnContinuousAnalyzer(column=col, nbins=None, yscale=auto, xmin=q0, xmax=q1, figsize=None)
      (transformer): ToNumericSeriesTransformer()
    )
    >>> analyzer = column.get_analyzer()
    >>> analyzer
    ColumnContinuousAnalyzer(column=col, nbins=None, yscale=auto, xmin=q0, xmax=q1, figsize=None)
    >>> transformer = column.get_transformer()
    >>> transformer
    ToNumericSeriesTransformer()

    ```
    """

    @abstractmethod
    def get_analyzer(self) -> BaseAnalyzer:
        r"""Get the column analyzer.

        Returns:
            The column analyzer.
        """

    @abstractmethod
    def get_transformer(self) -> BaseSeriesTransformer:
        r"""Get the column transformer.

        Returns:
            The column transformer.
        """


class Column(BaseColumn):
    r"""Define the column base class.

    Args:
        can_be_null: ``True`` if the column can have null values,
            otherwise ``False``.
        analyzer: The column analyzer or its configuration.
        transformer: The column  transformer or its
        `configuration.

    Example usage:

    ```pycon
    >>> from flamme.utils.columns import Column
    >>> from flamme.analyzer import ColumnContinuousAnalyzer
    >>> from flamme.transformer.series import ToNumeric
    >>> column = Column(
    ...     can_be_null=True,
    ...     analyzer=ColumnContinuousAnalyzer(column="col"),
    ...     transformer=ToNumeric(),
    ... )
    >>> column
    Column(
      (can_be_null): True
      (analyzer): ColumnContinuousAnalyzer(column=col, nbins=None, yscale=auto, xmin=q0, xmax=q1, figsize=None)
      (transformer): ToNumericSeriesTransformer()
    )
    >>> analyzer = column.get_analyzer()
    >>> analyzer
    ColumnContinuousAnalyzer(column=col, nbins=None, yscale=auto, xmin=q0, xmax=q1, figsize=None)
    >>> transformer = column.get_transformer()
    >>> transformer
    ToNumericSeriesTransformer()

    ```
    """

    def __init__(
        self,
        can_be_null: bool,
        analyzer: BaseAnalyzer | dict,
        transformer: BaseSeriesTransformer | dict,
    ) -> None:
        self._can_be_null = bool(can_be_null)
        self._analyzer = setup_analyzer(analyzer)
        self._transformer = setup_series_transformer(transformer)

    def __repr__(self) -> str:
        args = str_indent(
            str_mapping(
                {
                    "can_be_null": self._can_be_null,
                    "analyzer": self._analyzer,
                    "transformer": self._transformer,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    @property
    def can_be_null(self) -> bool:
        r"""``True`` if the column can have null values, otherwise
        ``False``."""
        return self._can_be_null

    def get_analyzer(self) -> BaseAnalyzer:
        return self._analyzer

    def get_transformer(self) -> BaseSeriesTransformer:
        return self._transformer



r"""Contain the implementation of a section to analyze the temporal
distribution of a column with discrete values."""

from __future__ import annotations

__all__ = ["ColumnTemporalDiscreteSection"]

import logging
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
from coola.utils import repr_indent, repr_mapping
from jinja2 import Template
from matplotlib import pyplot as plt

from flamme.section.base import BaseSection
from flamme.section.utils import (
    GO_TO_TOP,
    render_html_toc,
    tags2id,
    tags2title,
    valid_h_tag,
)
from flamme.utils.figure import figure2html, readable_xticklabels
from flamme.utils.sorting import mixed_typed_sort

if TYPE_CHECKING:
    from collections.abc import Sequence

logger = logging.getLogger(__name__)


class ColumnTemporalDiscreteSection(BaseSection):
    r"""Implement a section that analyzes the temporal distribution of a
    column with discrete values.

    Args:
        frame: The DataFrame to analyze.
        column: The column of the DataFrame to analyze.
        dt_column: The datetime column used to analyze
            the temporal distribution.
        period: The temporal period e.g. monthly or
            daily.
        figsize: The figure size in inches. The first
            dimension is the width and the second is the height.

    Example usage:

    ```pycon
    >>> import pandas as pd
    >>> import numpy as np
    >>> from flamme.section import ColumnTemporalDiscreteSection
    >>> section = ColumnTemporalDiscreteSection(
    ...     frame=pd.DataFrame(
    ...         {
    ...             "col": np.array([1, 42, np.nan, 22]),
    ...             "col2": ["a", "b", 1, "a"],
    ...             "datetime": pd.to_datetime(
    ...                 ["2020-01-03", "2020-02-03", "2020-03-03", "2020-04-03"]
    ...             ),
    ...         }
    ...     ),
    ...     column="col",
    ...     dt_column="datetime",
    ...     period="M",
    ... )
    >>> section
    ColumnTemporalDiscreteSection(
      (column): col
      (dt_column): datetime
      (period): M
      (figsize): None
    )
    >>> section.get_statistics()
    {}

    ```
    """

    def __init__(
        self,
        frame: pd.DataFrame,
        column: str,
        dt_column: str,
        period: str,
        figsize: tuple[float, float] | None = None,
    ) -> None:
        self._frame = frame
        self._column = column
        self._dt_column = dt_column
        self._period = period
        self._figsize = figsize

    def __repr__(self) -> str:
        args = repr_indent(
            repr_mapping(
                {
                    "column": self._column,
                    "dt_column": self._dt_column,
                    "period": self._period,
                    "figsize": self._figsize,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    @property
    def column(self) -> str:
        return self._column

    @property
    def dt_column(self) -> str:
        return self._dt_column

    @property
    def period(self) -> str:
        return self._period

    @property
    def figsize(self) -> tuple[float, float] | None:
        r"""The individual figure size in pixels.

        The first dimension is the width and the second is the height.
        """
        return self._figsize

    def get_statistics(self) -> dict:
        return {}

    def render_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        logger.info(
            f"Analyzing the temporal discrete distribution of {self._column} | "
            f"datetime column: {self._dt_column} | period: {self._period}"
        )
        return Template(self._create_template()).render(
            {
                "go_to_top": GO_TO_TOP,
                "id": tags2id(tags),
                "depth": valid_h_tag(depth + 1),
                "title": tags2title(tags),
                "section": number,
                "column": self._column,
                "dt_column": self._dt_column,
                "period": self._period,
                "figure": create_temporal_figure(
                    frame=self._frame,
                    column=self._column,
                    dt_column=self._dt_column,
                    period=self._period,
                    figsize=self._figsize,
                ),
            }
        )

    def render_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        return render_html_toc(number=number, tags=tags, depth=depth, max_depth=max_depth)

    def _create_template(self) -> str:
        return """
<h{{depth}} id="{{id}}">{{section}} {{title}} </h{{depth}}>

{{go_to_top}}

<p style="margin-top: 1rem;">
This section analyzes the temporal distribution of column <em>{{column}}</em>
by using the column <em>{{dt_column}}</em>.

{{figure}}

{{table}}
<p style="margin-top: 1rem;">
"""


def create_temporal_figure(
    frame: pd.DataFrame,
    column: str,
    dt_column: str,
    period: str,
    figsize: tuple[float, float] | None = None,
) -> str:
    r"""Create a HTML representation of a figure with the temporal value
    distribution.

    Args:
        frame: The DataFrame to analyze.
        column: The column to analyze.
        dt_column: The datetime column used to analyze
            the temporal distribution.
        period: The temporal period e.g. monthly or
            daily.
        log_y (bool, optional): If ``True``, it represents the bars
            with a log scale. Default: ``False``
        figsize: The figure size in inches. The first
            dimension is the width and the second is the height.

    Returns:
        The HTML representation of the figure.
    """
    if frame.shape[0] == 0:
        return "<span>&#9888;</span> No figure is generated because the column is empty"
    frame = frame[[column, dt_column]].copy()
    col_dt, col_count = "__datetime__", "__count__"
    frame[col_dt] = frame[dt_column].dt.to_period(period).astype(str)
    frame = frame[[column, col_dt]].groupby(by=[col_dt, column], dropna=False)[column].size()
    frame = pd.DataFrame({col_count: frame}).reset_index().sort_values(by=[col_dt, column])
    frame = frame.pivot_table(
        index=col_dt, columns=column, values=col_count, fill_value=0, dropna=False
    )

    labels = mixed_typed_sort(frame.columns.tolist())
    num_labels = len(labels)
    steps = frame.index.tolist()
    x = np.arange(len(steps), dtype=int)
    bottom = np.zeros_like(x)
    width = 0.9 if len(steps) < 50 else 1
    fig, ax = plt.subplots(figsize=figsize)
    my_cmap = plt.get_cmap("viridis")
    for i, label in enumerate(labels):
        count = frame[label].to_numpy().astype(int)
        ax.bar(x, count, label=label, bottom=bottom, width=width, color=my_cmap(i / num_labels))
        bottom += count

    if len(labels) <= 10:
        ax.legend()
    ax.set_xticks(x, labels=steps)
    readable_xticklabels(ax, max_num_xticks=100)
    ax.set_xlim(-0.5, len(steps) - 0.5)
    ax.set_ylabel("Number of occurrences")
    ax.set_title(f"Temporal distribution for column {column}")
    return figure2html(fig, close_fig=True)



r"""Contain reporters."""

from __future__ import annotations

__all__ = [
    "BaseReporter",
    "NoRepeatReporter",
    "Reporter",
    "is_reporter_config",
    "setup_reporter",
]

from flamme.reporter.base import BaseReporter, is_reporter_config, setup_reporter
from flamme.reporter.no_repeat import NoRepeatReporter
from flamme.reporter.vanilla import Reporter



r"""Contain ``pandas.DataFrame`` transformers to transform columns with
string values."""

from __future__ import annotations

__all__ = ["StripStringDataFrameTransformer"]

from typing import TYPE_CHECKING

from tqdm import tqdm

from flamme.transformer.dataframe.base import BaseDataFrameTransformer

if TYPE_CHECKING:
    from collections.abc import Sequence

    import pandas as pd


class StripStringDataFrameTransformer(BaseDataFrameTransformer):
    r"""Implement a transformer to strip the strings of some columns.

    Args:
        columns: The columns to process.

    Example usage:

    ```pycon
    >>> import pandas as pd
    >>> from flamme.transformer.dataframe import StripString
    >>> transformer = StripString(columns=["col1", "col3"])
    >>> transformer
    StripStringDataFrameTransformer(columns=('col1', 'col3'))
    >>> frame = pd.DataFrame(
    ...     {
    ...         "col1": [1, 2, 3, 4, 5],
    ...         "col2": ["1", "2", "3", "4", "5"],
    ...         "col3": ["a ", " b", "  c  ", "d", "e"],
    ...         "col4": ["a ", " b", "  c  ", "d", "e"],
    ...     }
    ... )
    >>> frame
       col1 col2   col3   col4
    0     1    1     a      a
    1     2    2      b      b
    2     3    3    c      c
    3     4    4      d      d
    4     5    5      e      e
    >>> out = transformer.transform(frame)
    >>> out
       col1 col2 col3   col4
    0     1    1    a     a
    1     2    2    b      b
    2     3    3    c    c
    3     4    4    d      d
    4     5    5    e      e

    ```
    """

    def __init__(self, columns: Sequence[str]) -> None:
        self._columns = tuple(columns)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(columns={self._columns})"

    def transform(self, frame: pd.DataFrame) -> pd.DataFrame:
        for col in tqdm(self._columns, desc="Striping strings"):
            frame[col] = frame[col].apply(lambda x: x.strip() if isinstance(x, str) else x)
        return frame



r"""Contain the implementation of sections to analyze the number null
values."""

from __future__ import annotations

__all__ = ["NullValueSection"]

import logging
from typing import TYPE_CHECKING

import pandas as pd
from coola.utils import repr_indent, repr_mapping
from jinja2 import Template
from matplotlib import pyplot as plt

from flamme.section.base import BaseSection
from flamme.section.utils import (
    GO_TO_TOP,
    render_html_toc,
    tags2id,
    tags2title,
    valid_h_tag,
)
from flamme.utils.figure import figure2html, readable_xticklabels

if TYPE_CHECKING:
    from collections.abc import Sequence

    import numpy as np


logger = logging.getLogger(__name__)


class NullValueSection(BaseSection):
    r"""Implement a section that analyzes the number of null values.

    Args:
        columns: The column names.
        null_count: The number of null values for each column.
        total_count: The total number of values for each column.
        figsize: The figure size in inches. The first dimension
            is the width and the second is the height.

    Example usage:

    ```pycon
    >>> import numpy as np
    >>> from flamme.section import NullValueSection
    >>> section = NullValueSection(
    ...     columns=["col1", "col2", "col3"],
    ...     null_count=np.array([0, 1, 2]),
    ...     total_count=np.array([5, 5, 5]),
    ... )
    >>> section
    NullValueSection(
      (columns): ('col1', 'col2', 'col3')
      (null_count): array([0, 1, 2])
      (total_count): array([5, 5, 5])
      (figsize): None
    )
    >>> section.get_statistics()
    {'columns': ('col1', 'col2', 'col3'), 'null_count': (0, 1, 2), 'total_count': (5, 5, 5)}

    ```
    """

    def __init__(
        self,
        columns: Sequence[str],
        null_count: np.ndarray,
        total_count: np.ndarray,
        figsize: tuple[float, float] | None = None,
    ) -> None:
        self._columns = tuple(columns)
        self._null_count = null_count.flatten().astype(int)
        self._total_count = total_count.flatten().astype(int)
        self._figsize = figsize

        if len(self._columns) != self._null_count.shape[0]:
            msg = (
                f"columns ({len(self._columns):,}) and null_count ({self._null_count.shape[0]:,}) "
                "do not match"
            )
            raise RuntimeError(msg)
        if len(self._columns) != self._total_count.shape[0]:
            msg = (
                f"columns ({len(self._columns):,}) and total_count "
                f"({self._total_count.shape[0]:,}) do not match"
            )
            raise RuntimeError(msg)

    def __repr__(self) -> str:
        args = repr_indent(
            repr_mapping(
                {
                    "columns": self._columns,
                    "null_count": self._null_count,
                    "total_count": self._total_count,
                    "figsize": self._figsize,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    @property
    def columns(self) -> tuple[str, ...]:
        r"""The columns used to compute the duplicated rows."""
        return self._columns

    @property
    def null_count(self) -> np.ndarray:
        r"""The number of null values for each column."""
        return self._null_count

    @property
    def total_count(self) -> np.ndarray:
        r"""The total number of values for each column."""
        return self._total_count

    @property
    def figsize(self) -> tuple[float, float] | None:
        r"""The individual figure size in pixels.

        The first dimension is the width and the second is the height.
        """
        return self._figsize

    def get_statistics(self) -> dict:
        return {
            "columns": self._columns,
            "null_count": tuple(self._null_count.tolist()),
            "total_count": tuple(self._total_count.tolist()),
        }

    def render_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        logger.info("Rendering the null value distribution of all columns...")
        return Template(self._create_template()).render(
            {
                "go_to_top": GO_TO_TOP,
                "id": tags2id(tags),
                "depth": valid_h_tag(depth + 1),
                "title": tags2title(tags),
                "section": number,
                "table_alpha": self._create_table(sort_by="column"),
                "table_sort": self._create_table(sort_by="null"),
                "bar_figure": self._create_bar_figure(),
                "num_columns": f"{len(self._columns):,}",
            }
        )

    def render_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        return render_html_toc(number=number, tags=tags, depth=depth, max_depth=max_depth)

    def _create_template(self) -> str:
        return """
<h{{depth}} id="{{id}}">{{section}} {{title}} </h{{depth}}>

{{go_to_top}}

<p style="margin-top: 1rem;">
This section analyzes the number and proportion of null values for the {{num_columns}}
columns.
In the following histogram, the columns are sorted by ascending order of null values.

{{bar_figure}}

<details>
    <summary>[show statistics per column]</summary>

    <p style="margin-top: 1rem;">
    The following tables show the number and proportion of null values for the {{num_columns}}
    columns.
    The background color of the row indicates the proportion of missing values:
    dark blues indicates more missing values than light blues.

    <div class="container-fluid">
        <div class="row align-items-start">
            <div class="col align-self-center">
                <p><b>Columns sorted by alphabetical order</b></p>

                {{table_alpha}}

            </div>
            <div class="col">
                <p><b>Columns sorted by ascending order of missing values</b></p>

                {{table_sort}}

            </div>
        </div>
    </div>
</details>

<p style="margin-top: 1rem;">
"""

    def _create_bar_figure(self) -> str:
        dataframe = self._get_dataframe().sort_values(by="null")
        fig, ax = plt.subplots(figsize=self._figsize)
        labels = dataframe["column"].tolist()
        ax.bar(x=labels, height=dataframe["null"].to_numpy(), color="tab:blue")
        if labels:
            ax.set_xlim(-0.5, len(labels) - 0.5)
        readable_xticklabels(ax, max_num_xticks=100)
        ax.set_xlabel("column")
        ax.set_ylabel("number of null values")
        ax.set_title("number of null values per column")
        return figure2html(fig, close_fig=True)

    def _create_table(self, sort_by: str) -> str:
        dataframe = self._get_dataframe().sort_values(by=sort_by)
        rows = "\n".join(
            [
                create_table_row(column=column, null_count=null_count, total_count=total_count)
                for column, null_count, total_count in zip(
                    dataframe["column"].to_numpy(),
                    dataframe["null"].to_numpy(),
                    dataframe["total"].to_numpy(),
                )
            ]
        )
        return Template(
            """
<table class="table table-hover table-responsive w-auto" >
    <thead class="thead table-group-divider">
        <tr>
            <th>column</th>
            <th>null pct</th>
            <th>null count</th>
            <th>total count</th>
        </tr>
    </thead>
    <tbody class="tbody table-group-divider">
        {{rows}}
        <tr class="table-group-divider"></tr>
    </tbody>
</table>
"""
        ).render({"rows": rows})

    def _get_dataframe(self) -> pd.DataFrame:
        return pd.DataFrame(
            {"column": self._columns, "null": self._null_count, "total": self._total_count}
        )


def create_table_row(column: str, null_count: int, total_count: int) -> str:
    r"""Create the HTML code of a new table row.

    Args:
        column: The column name.
        null_count (int): The number of null values.
        total_count (int): The total number of rows.

    Returns:
        The HTML code of a row.
    """
    pct = null_count / total_count
    return Template(
        """<tr>
    <th style="background-color: rgba(0, 191, 255, {{null_pct}})">{{column}}</th>
    <td {{num_style}}>{{null_pct}}</td>
    <td {{num_style}}>{{null_count}}</td>
    <td {{num_style}}>{{total_count}}</td>
</tr>"""
    ).render(
        {
            "num_style": f'style="text-align: right; background-color: rgba(0, 191, 255, {pct})"',
            "column": column,
            "null_count": f"{null_count:,}",
            "null_pct": f"{pct:.4f}",
            "total_count": f"{total_count:,}",
        }
    )



r"""Contain the implementation of a section that generates the given
custom content."""

from __future__ import annotations

__all__ = ["ContentSection"]

import logging
from typing import TYPE_CHECKING

from jinja2 import Template

from flamme.section.base import BaseSection
from flamme.section.utils import (
    GO_TO_TOP,
    render_html_toc,
    tags2id,
    tags2title,
    valid_h_tag,
)

if TYPE_CHECKING:
    from collections.abc import Sequence


logger = logging.getLogger(__name__)


class ContentSection(BaseSection):
    r"""Implement a section that generates the given custom content.

    Args:
        content: The content to use in the HTML code.

    Example usage:

    ```pycon
    >>> from flamme.section import ContentSection
    >>> section = ContentSection(content="meow")
    >>> section
    ContentSection()
    >>> section.get_statistics()
    {}

    ```
    """

    def __init__(self, content: str) -> None:
        self._content = str(content)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def get_statistics(self) -> dict:
        return {}

    def render_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        logger.info("Rendering the section with the custom content...")
        return Template(self._create_template()).render(
            {
                "go_to_top": GO_TO_TOP,
                "id": tags2id(tags),
                "depth": valid_h_tag(depth + 1),
                "title": tags2title(tags),
                "section": number,
                "content": self._content,
            }
        )

    def render_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        return render_html_toc(number=number, tags=tags, depth=depth, max_depth=max_depth)

    def _create_template(self) -> str:
        return """
<h{{depth}} id="{{id}}">{{section}} {{title}} </h{{depth}}>

{{go_to_top}}

<p style="margin-top: 1rem;">

{{content}}

<p style="margin-top: 1rem;">
"""



r"""Contain the implementation of a section to analyze a column with
continuous values."""

from __future__ import annotations

__all__ = ["ColumnContinuousSection"]

import logging
from typing import TYPE_CHECKING

import numpy as np
from coola.utils import repr_indent, repr_mapping
from jinja2 import Template
from matplotlib import pyplot as plt
from matplotlib.lines import Line2D

from flamme.section.base import BaseSection
from flamme.section.utils import (
    GO_TO_TOP,
    auto_yscale_continuous,
    compute_statistics,
    render_html_toc,
    tags2id,
    tags2title,
    valid_h_tag,
)
from flamme.utils.array import nonnan
from flamme.utils.figure import figure2html, readable_xticklabels
from flamme.utils.range import find_range

if TYPE_CHECKING:
    from collections.abc import Sequence

    import pandas as pd
    from matplotlib.axes import Axes


logger = logging.getLogger(__name__)


class ColumnContinuousSection(BaseSection):
    r"""Implement a section that analyzes a continuous distribution of
    values.

    Args:
        series: The series/column to analyze.
        column: The column name.
        nbins: The number of bins in the histogram.
        yscale: The y-axis scale. If ``'auto'``, the
            ``'linear'`` or ``'log'/'symlog'`` scale is chosen based
            on the distribution.
        xmin: The minimum value of the range or its
            associated quantile. ``q0.1`` means the 10% quantile.
            ``0`` is the minimum value and ``1`` is the maximum value.
        xmax: The maximum value of the range or its
            associated quantile. ``q0.9`` means the 90% quantile.
            ``0`` is the minimum value and ``1`` is the maximum value.
        figsize: The figure size in inches. The first
            dimension is the width and the second is the height.

    Example usage:

    ```pycon
    >>> import pandas as pd
    >>> from flamme.section import ColumnContinuousSection
    >>> section = ColumnContinuousSection(
    ...     series=pd.Series([np.nan, *list(range(101)), np.nan]), column="col"
    ... )
    >>> section
    ColumnContinuousSection(
      (column): col
      (nbins): None
      (yscale): auto
      (xmin): None
      (xmax): None
      (figsize): None
    )
    >>> section.get_statistics()
    {'count': 103, 'num_nulls': 2, 'nunique': 102, 'mean': 50.0, 'std': 29.30...,
     'skewness': 0.0, 'kurtosis': -1.200235294117647, 'min': 0.0,
     'q001': 0.1, 'q01': 1.0, 'q05': 5.0, 'q10': 10.0, 'q25': 25.0, 'median': 50.0,
     'q75': 75.0, 'q90': 90.0, 'q95': 95.0, 'q99': 99.0, 'q999': 99.9, 'max': 100.0,
     '>0': 100, '<0': 0, '=0': 1, 'num_non_nulls': 101}

    ```
    """

    def __init__(
        self,
        series: pd.Series,
        column: str,
        nbins: int | None = None,
        yscale: str = "auto",
        xmin: float | str | None = None,
        xmax: float | str | None = None,
        figsize: tuple[float, float] | None = None,
    ) -> None:
        self._series = series
        self._column = column
        self._nbins = nbins
        self._yscale = yscale
        self._xmin = xmin
        self._xmax = xmax
        self._figsize = figsize

    def __repr__(self) -> str:
        args = repr_indent(
            repr_mapping(
                {
                    "column": self._column,
                    "nbins": self._nbins,
                    "yscale": self._yscale,
                    "xmin": self._xmin,
                    "xmax": self._xmax,
                    "figsize": self._figsize,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    @property
    def column(self) -> str:
        return self._column

    @property
    def yscale(self) -> str:
        return self._yscale

    @property
    def nbins(self) -> int | None:
        return self._nbins

    @property
    def series(self) -> pd.Series:
        return self._series

    @property
    def xmin(self) -> float | str | None:
        return self._xmin

    @property
    def xmax(self) -> float | str | None:
        return self._xmax

    @property
    def figsize(self) -> tuple[float, float] | None:
        r"""The individual figure size in pixels.

        The first dimension is the width and the second is the height.
        """
        return self._figsize

    def get_statistics(self) -> dict[str, float]:
        return compute_statistics(self._series)

    def render_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        logger.info(f"Rendering the continuous distribution of {self._column}")
        stats = self.get_statistics()
        null_values_pct = (
            f"{100 * stats['num_nulls'] / stats['count']:.2f}" if stats["count"] > 0 else "N/A"
        )
        return Template(self._create_template()).render(
            {
                "go_to_top": GO_TO_TOP,
                "id": tags2id(tags),
                "depth": valid_h_tag(depth + 1),
                "title": tags2title(tags),
                "section": number,
                "column": self._column,
                "table": create_stats_table(stats=stats, column=self._column),
                "total_values": f"{stats['count']:,}",
                "unique_values": f"{stats['nunique']:,}",
                "null_values": f"{stats['num_nulls']:,}",
                "null_values_pct": null_values_pct,
                "histogram_figure": create_histogram_figure(
                    series=self._series,
                    column=self._column,
                    stats=stats,
                    nbins=self._nbins,
                    yscale=self._yscale,
                    xmin=self._xmin,
                    xmax=self._xmax,
                    figsize=self._figsize,
                ),
                "boxplot_figure": create_boxplot_figure(
                    series=self._series,
                    xmin=self._xmin,
                    xmax=self._xmax,
                    figsize=self._figsize,
                ),
            }
        )

    def render_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        return render_html_toc(number=number, tags=tags, depth=depth, max_depth=max_depth)

    def _create_template(self) -> str:
        return """
<h{{depth}} id="{{id}}">{{section}} {{title}} </h{{depth}}>

{{go_to_top}}

<p style="margin-top: 1rem;">
This section analyzes the discrete distribution of values for column <em>{{column}}</em>.

<ul>
  <li> total values: {{total_values}} </li>
  <li> number of unique values: {{unique_values}} </li>
  <li> number of null values: {{null_values}} / {{total_values}} ({{null_values_pct}}%) </li>
</ul>

{{histogram_figure}}
{{boxplot_figure}}
{{table}}
<p style="margin-top: 1rem;">
"""


def create_boxplot_figure(
    series: pd.Series,
    xmin: float | str | None = None,
    xmax: float | str | None = None,
    figsize: tuple[float, float] | None = None,
) -> str:
    r"""Create the HTML code of a boxplot figure.

    Args:
        series: The series/column to analyze.
        xmin: The minimum value of the range or its
            associated quantile. ``q0.1`` means the 10% quantile.
            ``0`` is the minimum value and ``1`` is the maximum value.
        xmax: The maximum value of the range or its
            associated quantile. ``q0.9`` means the 90% quantile.
            ``0`` is the minimum value and ``1`` is the maximum value.
        figsize: The figure size in inches. The first
            dimension is the width and the second is the height.

    Returns:
        The HTML code of the figure.
    """
    array = series.dropna().to_numpy()
    if array.size == 0:
        return "<span>&#9888;</span> No figure is generated because the column is empty"
    xmin, xmax = find_range(array, xmin=xmin, xmax=xmax)
    if figsize is not None:
        figsize = (figsize[0], figsize[0] / 10)
    fig, ax = plt.subplots(figsize=figsize)
    ax.boxplot(
        array,
        notch=True,
        vert=False,
        widths=0.7,
        patch_artist=True,
        boxprops={"facecolor": "lightblue"},
    )
    readable_xticklabels(ax, max_num_xticks=100)
    if xmin < xmax:
        ax.set_xlim(xmin, xmax)
    ax.set_ylabel(" ")
    return figure2html(fig, close_fig=True)


def create_histogram_figure(
    series: pd.Series,
    column: str,
    stats: dict,
    nbins: int | None = None,
    yscale: str = "linear",
    xmin: float | str | None = None,
    xmax: float | str | None = None,
    figsize: tuple[float, float] | None = None,
) -> str:
    r"""Create the HTML code of a histogram figure.

    Args:
        series: The series/column to analyze.
        column: The column name.
        stats: Specifies a dictionary with the statistics.
        nbins: The number of bins in the histogram.
        yscale: The y-axis scale. If ``'auto'``, the
            ``'linear'`` or ``'log'`` scale is chosen based on the
            distribution.
        xmin: The minimum value of the range or its
            associated quantile. ``q0.1`` means the 10% quantile.
            ``0`` is the minimum value and ``1`` is the maximum value.
        xmax: The maximum value of the range or its
            associated quantile. ``q0.9`` means the 90% quantile.
            ``0`` is the minimum value and ``1`` is the maximum value.
        figsize: The figure size in inches. The first
            dimension is the width and the second is the height.

    Returns:
        The HTML code of the figure.
    """
    array = series.to_numpy()
    if array.size == 0:
        return "<span>&#9888;</span> No figure is generated because the column is empty"
    xmin, xmax = find_range(array, xmin=xmin, xmax=xmax)
    fig, ax = plt.subplots(figsize=figsize)
    ax.hist(array, bins=nbins, range=[xmin, xmax], color="tab:blue", alpha=0.9)
    readable_xticklabels(ax, max_num_xticks=100)
    if xmin < xmax:
        ax.set_xlim(xmin, xmax)
    ax.set_title(f"Distribution of values for column {column}")
    ax.set_ylabel("Number of occurrences")
    if yscale == "auto":
        yscale = auto_yscale_continuous(array=array, nbins=nbins)
    ax.set_yscale(yscale)
    if xmin < stats["q05"] < xmax:
        add_axvline_quantile(ax, x=stats["q05"], label="q0.05 ", horizontalalignment="right")
        ax.axvline(stats["q05"], color="black", linestyle="dashed")
    if xmin < stats["q95"] < xmax:
        add_axvline_quantile(ax, x=stats["q95"], label=" q0.95", horizontalalignment="left")
    # if xmin < stats["median"] < xmax:
    #     add_axvline_median(ax, x=stats["median"])
    add_cdf_plot(ax, array=array, nbins=nbins, xmin=xmin, xmax=xmax)
    ax.legend(
        [Line2D([0], [0], linestyle="none", mfc="black", mec="none", marker="")] * 3,
        [
            f'std={stats["std"]:.2f}',
            f'skewness={stats["skewness"]:.2f}',
            f'kurtosis={stats["kurtosis"]:.2f}',
        ],
    )
    return figure2html(fig, close_fig=True)


def add_axvline_quantile(
    ax: Axes,
    x: float,
    label: str,
    color: str = "black",
    linestyle: str = "dashed",
    horizontalalignment: str = "center",
) -> None:
    ax.axvline(x, color=color, linestyle=linestyle)
    ax.text(
        x,
        0.99,
        label,
        transform=ax.get_xaxis_transform(),
        color=color,
        horizontalalignment=horizontalalignment,
        verticalalignment="top",
    )


def add_axvline_median(
    ax: Axes,
    x: float,
    label: str = "median",
    color: str = "black",
    linestyle: str = "dashed",
    horizontalalignment: str = "center",
) -> None:
    ax.axvline(x, color=color, linestyle=linestyle)
    ax.text(
        x,
        0.99,
        label,
        transform=ax.get_xaxis_transform(),
        color=color,
        horizontalalignment=horizontalalignment,
        verticalalignment="top",
    )


def add_cdf_plot(
    ax: Axes,
    array: np.ndarray,
    nbins: int,
    xmin: float = float("-inf"),
    xmax: float = float("inf"),
    color: str = "tab:red",
) -> None:
    r"""Add the cumulative distribution function (CDF) plot.

    Args:
        ax: The axes of the matplotlib figure to update.
        array: The array with the data.
        nbins: The number of bins to use to plot the CDF.
        xmin: The minimum value of the range or its
            associated quantile. ``q0.1`` means the 10% quantile.
            ``0`` is the minimum value and ``1`` is the maximum value.
        xmax: The maximum value of the range or its
            associated quantile. ``q0.9`` means the 90% quantile.
            ``0`` is the minimum value and ``1`` is the maximum value.
        color: The plot color.
    """
    nbins = nbins or 1000
    array = nonnan(array)
    if array.size == 0:
        return
    nleft = array[array < xmin].size
    nright = array[array > xmax].size
    counts, edges = np.histogram(array[np.logical_and(array >= xmin, array <= xmax)], bins=nbins)
    cdf = (np.cumsum(counts) + nleft) / (np.sum(counts) + nleft + nright)
    x = [(left + right) * 0.5 for left, right in zip(edges[:-1], edges[1:])]
    ax = ax.twinx()
    ax.tick_params(axis="y", labelcolor=color)
    ax.plot(x, cdf, color=color, label="CDF")
    xmin, xmax = max(0.0, cdf[0]), min(1.0, cdf[-1])
    if xmin < xmax:
        ax.set_ylim(xmin, xmax)
    ax.set_ylabel("cumulative distribution function (CDF)", color=color)


def create_stats_table(stats: dict, column: str) -> str:
    r"""Create the HTML code of the table with statistics.

    Args:
        stats: Specifies a dictionary with the statistics.
        column: The column name.

    Returns:
        The HTML code of the table.
    """
    return Template(
        """
<details>
    <summary>[show statistics]</summary>

    <p>The following table shows some statistics about the distribution for column {{column}}.

    <table class="table table-hover table-responsive w-auto" >
        <thead class="thead table-group-divider">
            <tr><th>stat</th><th>value</th></tr>
        </thead>
        <tbody class="tbody table-group-divider">
            <tr><th>count</th><td {{num_style}}>{{count}}</td></tr>
            <tr><th>mean</th><td {{num_style}}>{{mean}}</td></tr>
            <tr><th>std</th><td {{num_style}}>{{std}}</td></tr>
            <tr><th>skewness</th><td {{num_style}}>{{skewness}}</td></tr>
            <tr><th>kurtosis</th><td {{num_style}}>{{kurtosis}}</td></tr>
            <tr><th>min</th><td {{num_style}}>{{min}}</td></tr>
            <tr><th>quantile 0.1%</th><td {{num_style}}>{{q01}}</td></tr>
            <tr><th>quantile 1%</th><td {{num_style}}>{{q01}}</td></tr>
            <tr><th>quantile 5%</th><td {{num_style}}>{{q05}}</td></tr>
            <tr><th>quantile 10%</th><td {{num_style}}>{{q10}}</td></tr>
            <tr><th>quantile 25%</th><td {{num_style}}>{{q25}}</td></tr>
            <tr><th>median</th><td {{num_style}}>{{median}}</td></tr>
            <tr><th>quantile 75%</th><td {{num_style}}>{{q75}}</td></tr>
            <tr><th>quantile 90%</th><td {{num_style}}>{{q90}}</td></tr>
            <tr><th>quantile 95%</th><td {{num_style}}>{{q95}}</td></tr>
            <tr><th>quantile 99%</th><td {{num_style}}>{{q99}}</td></tr>
            <tr><th>quantile 99.9%</th><td {{num_style}}>{{q99}}</td></tr>
            <tr><th>max</th><td {{num_style}}>{{max}}</td></tr>
            <tr><th>number of zeros</th><td {{num_style}}>{{num_zeros}}</td></tr>
            <tr><th>number of positive values</th><td {{num_style}}>{{num_pos}}</td></tr>
            <tr><th>number of negative values</th><td {{num_style}}>{{num_neg}}</td></tr>
            <tr class="table-group-divider"></tr>
        </tbody>
    </table>
</details>
"""
    ).render(
        {
            "column": column,
            "num_style": 'style="text-align: right;"',
            "count": f"{stats['count']:,}",
            "mean": f"{stats['mean']:,.4f}",
            "std": f"{stats['std']:,.4f}",
            "skewness": f"{stats['skewness']:,.4f}",
            "kurtosis": f"{stats['kurtosis']:,.4f}",
            "min": f"{stats['min']:,.4f}",
            "q001": f"{stats['q001']:,.4f}",
            "q01": f"{stats['q01']:,.4f}",
            "q05": f"{stats['q05']:,.4f}",
            "q10": f"{stats['q10']:,.4f}",
            "q25": f"{stats['q25']:,.4f}",
            "median": f"{stats['median']:,.4f}",
            "q75": f"{stats['q75']:,.4f}",
            "q90": f"{stats['q90']:,.4f}",
            "q95": f"{stats['q95']:,.4f}",
            "q99": f"{stats['q99']:,.4f}",
            "q999": f"{stats['q999']:,.4f}",
            "max": f"{stats['max']:,.4f}",
            "num_pos": f"{stats['>0']:,}",
            "num_neg": f"{stats['<0']:,}",
            "num_zeros": f"{stats['=0']:,}",
        }
    )



r"""Implement continuous values analyzers."""

from __future__ import annotations

__all__ = ["ColumnContinuousAdvancedAnalyzer"]

import logging
from typing import TYPE_CHECKING

from flamme.analyzer.base import BaseAnalyzer
from flamme.section import ColumnContinuousAdvancedSection, EmptySection

if TYPE_CHECKING:
    import pandas as pd

logger = logging.getLogger(__name__)


class ColumnContinuousAdvancedAnalyzer(BaseAnalyzer):
    r"""Implement an analyzer to show the temporal distribution of
    continuous values.

    Args:
        column: The column name.
        nbins: The number of bins in the histogram.
        yscale: The y-axis scale. If ``'auto'``, the
            ``'linear'`` or ``'log'/'symlog'`` scale is chosen based
            on the distribution.
        figsize: The figure size in inches. The first
            dimension is the width and the second is the height.

    Example usage:

    ```pycon
    >>> import numpy as np
    >>> import pandas as pd
    >>> from flamme.analyzer import ColumnContinuousAdvancedAnalyzer
    >>> analyzer = ColumnContinuousAdvancedAnalyzer(column="float")
    >>> analyzer
    ColumnContinuousAdvancedAnalyzer(column=float, nbins=None, yscale=auto, figsize=None)
    >>> frame = pd.DataFrame(
    ...     {
    ...         "int": np.array([np.nan, 1, 0, 1]),
    ...         "float": np.array([1.2, 4.2, np.nan, 2.2]),
    ...         "str": np.array(["A", "B", None, np.nan]),
    ...     }
    ... )
    >>> section = analyzer.analyze(frame)

    ```
    """

    def __init__(
        self,
        column: str,
        nbins: int | None = None,
        yscale: str = "auto",
        figsize: tuple[float, float] | None = None,
    ) -> None:
        self._column = column
        self._nbins = nbins
        self._yscale = yscale
        self._figsize = figsize

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(column={self._column}, nbins={self._nbins}, "
            f"yscale={self._yscale}, figsize={self._figsize})"
        )

    def analyze(self, frame: pd.DataFrame) -> ColumnContinuousAdvancedSection | EmptySection:
        logger.info(f"Analyzing the continuous distribution of {self._column}")
        if self._column not in frame:
            logger.info(
                "Skipping temporal continuous distribution analysis because the column "
                f"({self._column}) is not in the DataFrame: {sorted(frame.columns)}"
            )
            return EmptySection()
        return ColumnContinuousAdvancedSection(
            column=self._column,
            series=frame[self._column],
            nbins=self._nbins,
            yscale=self._yscale,
            figsize=self._figsize,
        )



r"""Contain utility functions to format strings."""

from __future__ import annotations

__all__ = ["human_byte"]


def human_byte(size: float, decimal: int = 2) -> str:
    r"""Return a human-readable string representation of byte sizes.

    Args:
        size: The number of bytes.
        decimal: The number of decimal digits.

    Returns:
        The human-readable string representation of byte sizes.

    Example usage:

    ```pycon
    >>> from flamme.utils.format import human_byte
    >>> human_byte(2)
    '2.00 B'
    >>> human_byte(2048)
    '2.00 KB'
    >>> human_byte(2097152)
    '2.00 MB'

    ```
    """
    for unit in ["B", "KB", "MB", "GB", "TB", "PB"]:
        if size < 1024.0:
            break
        if unit != "PB":
            size /= 1024.0
    return f"{size:,.{decimal}f} {unit}"



r"""Implement an analyzer that filters the data before to analyze the
data."""

from __future__ import annotations

__all__ = ["FilteredAnalyzer"]

import logging
from typing import TYPE_CHECKING

from coola.utils import str_indent, str_mapping

from flamme.analyzer.base import BaseAnalyzer, setup_analyzer

if TYPE_CHECKING:

    import pandas as pd

    from flamme.section import BaseSection

logger = logging.getLogger(__name__)


class FilteredAnalyzer(BaseAnalyzer):
    r"""Implement an analyzer that filters the data before to analyze the
    data.

    Args:
        query: Soecifies the query.
        analyzer: The analyzer or its configuration.

    Example usage:

    ```pycon
    >>> import numpy as np
    >>> import pandas as pd
    >>> from flamme.analyzer import FilteredAnalyzer, NullValueAnalyzer
    >>> analyzer = FilteredAnalyzer(query="float >= 2.0", analyzer=NullValueAnalyzer())
    >>> analyzer
    FilteredAnalyzer(
      (query): float >= 2.0
      (analyzer): NullValueAnalyzer(figsize=None)
    )
    >>> frame = pd.DataFrame(
    ...     {
    ...         "int": np.array([np.nan, 1, 0, 1]),
    ...         "float": np.array([1.2, 4.2, np.nan, 2.2]),
    ...         "str": np.array(["A", "B", None, np.nan]),
    ...     }
    ... )
    >>> section = analyzer.analyze(frame)

    ```
    """

    def __init__(self, query: str, analyzer: BaseAnalyzer | dict) -> None:
        self._query = query
        self._analyzer = setup_analyzer(analyzer)

    def __repr__(self) -> str:
        args = str_indent(str_mapping({"query": self._query, "analyzer": self._analyzer}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def analyze(self, frame: pd.DataFrame) -> BaseSection:
        logger.info("Filtering the DataFrame...")
        frame = frame.query(self._query)
        return self._analyzer.analyze(frame)



r"""Contain a wrapper around an ingestor to transform the data after
ingestion."""

from __future__ import annotations

__all__ = ["TransformedIngestor"]

import logging
from typing import TYPE_CHECKING

from coola.utils import str_indent, str_mapping

from flamme.ingestor.base import BaseIngestor, setup_ingestor
from flamme.transformer.dataframe.base import (
    BaseDataFrameTransformer,
    setup_dataframe_transformer,
)

if TYPE_CHECKING:
    import pandas as pd

logger = logging.getLogger(__name__)


class TransformedIngestor(BaseIngestor):
    r"""Implement an ingestor that also transforms the DataFrame.

    Args:
        ingestor: The base ingestor.
        transformer: Specifies a ``pandas.DataFrame`` transformer or
            its configuration.

    Example usage:

    ```pycon
    >>> from flamme.ingestor import TransformedIngestor, ParquetIngestor
    >>> from flamme.transformer.dataframe import ToNumeric
    >>> ingestor = TransformedIngestor(
    ...     ingestor=ParquetIngestor(path="/path/to/frame.csv"),
    ...     transformer=ToNumeric(columns=["col1", "col3"]),
    ... )
    >>> ingestor
    TransformedIngestor(
      (ingestor): ParquetIngestor(path=/path/to/frame.csv)
      (transformer): ToNumericDataFrameTransformer(columns=('col1', 'col3'))
    )
    >>> frame = ingestor.ingest()  # doctest: +SKIP

    ```
    """

    def __init__(
        self, ingestor: BaseIngestor | dict, transformer: BaseDataFrameTransformer | dict
    ) -> None:
        self._ingestor = setup_ingestor(ingestor)
        self._transformer = setup_dataframe_transformer(transformer)

    def __repr__(self) -> str:
        args = str_indent(
            str_mapping({"ingestor": self._ingestor, "transformer": self._transformer})
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def ingest(self) -> pd.DataFrame:
        frame = self._ingestor.ingest()
        return self._transformer.transform(frame)



r"""Implement an analyzer that generates a section about the most
frequent values in a given columns."""

from __future__ import annotations

__all__ = ["MostFrequentValuesAnalyzer"]

import logging
from collections import Counter
from typing import TYPE_CHECKING

from flamme.analyzer.base import BaseAnalyzer
from flamme.section import EmptySection, MostFrequentValuesSection

if TYPE_CHECKING:
    import pandas as pd

logger = logging.getLogger(__name__)


class MostFrequentValuesAnalyzer(BaseAnalyzer):
    r"""Implement a most frequent values analyzer for a given column.

    Args:
        column: The column to analyze.
        dropna: If ``True``, the NaN values are not included in the
            analysis.
        top: The maximum number of values to show.

    Example usage:

    ```pycon
    >>> import numpy as np
    >>> import pandas as pd
    >>> from flamme.analyzer import MostFrequentValuesAnalyzer
    >>> analyzer = MostFrequentValuesAnalyzer(column="str")
    >>> analyzer
    MostFrequentValuesAnalyzer(column=str, dropna=False, top=100)
    >>> frame = pd.DataFrame({"col": np.array([np.nan, 1, 0, 1])})
    >>> section = analyzer.analyze(frame)

    ```
    """

    def __init__(self, column: str, dropna: bool = False, top: int = 100) -> None:
        self._column = column
        self._dropna = bool(dropna)
        self._top = top

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(column={self._column}, "
            f"dropna={self._dropna}, top={self._top:,})"
        )

    def analyze(self, frame: pd.DataFrame) -> MostFrequentValuesSection | EmptySection:
        logger.info(f"Analyzing the most frequent values of {self._column}")
        if self._column not in frame:
            logger.info(
                f"Skipping most frequent values analysis of column {self._column} "
                f"because it is not in the DataFrame: {sorted(frame.columns)}"
            )
            return EmptySection()
        return MostFrequentValuesSection(
            counter=Counter(frame[self._column].value_counts(dropna=self._dropna).to_dict()),
            column=self._column,
            top=self._top,
        )



r"""Contain ``pandas.Series`` transformers to transform datetime
columns."""

from __future__ import annotations

__all__ = ["ToDatetimeSeriesTransformer"]

from typing import Any

import pandas as pd

from flamme.transformer.series.base import BaseSeriesTransformer


class ToDatetimeSeriesTransformer(BaseSeriesTransformer):
    r"""Implement a transformer to convert a ``pandas.Series`` to
    datetime type.

    Args:
        **kwargs: The keyword arguments for
            ``pandas.to_datetime``.

    Example usage:

    ```pycon
    >>> import pandas as pd
    >>> from flamme.transformer.series import ToDatetime
    >>> transformer = ToDatetime()
    >>> transformer
    ToDatetimeSeriesTransformer()
    >>> series = pd.Series(["2020-1-1", "2020-1-2", "2020-1-31", "2020-12-31", "2021-12-31"])
    >>> series.dtype
    dtype('O')
    >>> out = transformer.transform(series)
    >>> out.dtype
    dtype('<M8[ns]')

    ```
    """

    def __init__(self, **kwargs: Any) -> None:
        self._kwargs = kwargs

    def __repr__(self) -> str:
        args = ", ".join([f"{key}={value}" for key, value in self._kwargs.items()])
        return f"{self.__class__.__qualname__}({args})"

    def transform(self, series: pd.Series) -> pd.Series:
        return pd.to_datetime(series, **self._kwargs)



r"""Contain the implementation of sections to analyze the number null
values for each column."""

from __future__ import annotations

__all__ = ["ColumnTemporalNullValueSection"]

import logging
from typing import TYPE_CHECKING

from coola.utils import repr_indent, repr_mapping, str_indent
from jinja2 import Template
from matplotlib import pyplot as plt
from tqdm import tqdm

from flamme.section.base import BaseSection
from flamme.section.null_temp import plot_temporal_null_total
from flamme.section.utils import (
    GO_TO_TOP,
    render_html_toc,
    tags2id,
    tags2title,
    valid_h_tag,
)
from flamme.utils.figure import figure2html, readable_xticklabels

if TYPE_CHECKING:
    from collections.abc import Sequence

    import numpy as np
    import pandas as pd

logger = logging.getLogger(__name__)


class ColumnTemporalNullValueSection(BaseSection):
    r"""Implement a section to analyze the temporal distribution of null
    values for all columns.

    Args:
        frame: The DataFrame to analyze.
        columns: The list of columns to analyze. A plot is generated
            for each column.
        dt_column: The datetime column used to analyze
            the temporal distribution.
        period: The temporal period e.g. monthly or daily.
        ncols: The number of columns.
        figsize: The figure size in inches. The first dimension
            is the width and the second is the height.

    Example usage:

    ```pycon
    >>> import pandas as pd
    >>> import numpy as np
    >>> from flamme.section import ColumnTemporalNullValueSection
    >>> dataframe = pd.DataFrame(
    ...     {
    ...         "float": np.array([1.2, 4.2, np.nan, 2.2]),
    ...         "int": np.array([np.nan, 1, 0, 1]),
    ...         "str": np.array(["A", "B", None, np.nan]),
    ...         "datetime": pd.to_datetime(
    ...             ["2020-01-03", "2020-02-03", "2020-03-03", "2020-04-03"]
    ...         ),
    ...     }
    ... )
    >>> section = ColumnTemporalNullValueSection(
    ...     frame=dataframe, columns=["float", "int", "str"], dt_column="datetime", period="M"
    ... )
    >>> section
    ColumnTemporalNullValueSection(
      (columns): ('float', 'int', 'str')
      (dt_column): datetime
      (period): M
      (ncols): 2
      (figsize): (7, 5)
    )
    >>> section.get_statistics()
    {}

    ```
    """

    def __init__(
        self,
        frame: pd.DataFrame,
        columns: Sequence[str],
        dt_column: str,
        period: str,
        ncols: int = 2,
        figsize: tuple[float, float] = (7, 5),
    ) -> None:
        self._frame = frame
        self._columns = tuple(columns)
        self._dt_column = dt_column
        self._period = period
        self._ncols = min(ncols, len(self._columns))
        self._figsize = figsize

    def __repr__(self) -> str:
        args = repr_indent(
            repr_mapping(
                {
                    "columns": self._columns,
                    "dt_column": self._dt_column,
                    "period": self._period,
                    "ncols": self._ncols,
                    "figsize": self._figsize,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    @property
    def frame(self) -> pd.DataFrame:
        r"""The DataFrame to analyze."""
        return self._frame

    @property
    def columns(self) -> tuple[str, ...]:
        r"""The columns to analyze."""
        return self._columns

    @property
    def dt_column(self) -> str:
        r"""The datetime column."""
        return self._dt_column

    @property
    def period(self) -> str:
        r"""The temporal period used to analyze the data."""
        return self._period

    @property
    def ncols(self) -> int:
        r"""The number of columns to show the figures."""
        return self._ncols

    @property
    def figsize(self) -> tuple[float, float] | None:
        r"""The individual figure size in pixels.

        The first dimension is the width and the second is the height.
        """
        return self._figsize

    def get_statistics(self) -> dict:
        return {}

    def render_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        logger.info(
            f"Rendering the temporal null value distribution of the following columns: "
            f"{self._columns}\ndatetime column: {self._dt_column} | period: {self._period}"
        )
        return Template(self._create_template()).render(
            {
                "go_to_top": GO_TO_TOP,
                "id": tags2id(tags),
                "depth": valid_h_tag(depth + 1),
                "title": tags2title(tags),
                "section": number,
                "column": self._dt_column,
                "figure": create_temporal_null_figure(
                    frame=self._frame,
                    columns=self._columns,
                    dt_column=self._dt_column,
                    period=self._period,
                    ncols=self._ncols,
                    figsize=self._figsize,
                ),
                "table": create_table_section(
                    frame=self._frame,
                    columns=self._columns,
                    dt_column=self._dt_column,
                    period=self._period,
                ),
            }
        )

    def render_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        return render_html_toc(number=number, tags=tags, depth=depth, max_depth=max_depth)

    def _create_template(self) -> str:
        return """
<h{{depth}} id="{{id}}">{{section}} {{title}} </h{{depth}}>

{{go_to_top}}

<p style="margin-top: 1rem;">
This section analyzes the temporal distribution of null values.
The column <em>{{column}}</em> is used as the temporal column.

{{figure}}

<p style="margin-top: 1rem;">

{{table}}

<p style="margin-top: 1rem;">
"""


def create_temporal_null_figure(
    frame: pd.DataFrame,
    columns: Sequence[str],
    dt_column: str,
    period: str,
    ncols: int = 2,
    figsize: tuple[float, float] = (7, 5),
) -> str:
    r"""Create a HTML representation of a figure with the temporal null
    value distribution.

    Args:
        frame: The DataFrame to analyze.
        columns: The list of columns to analyze. A plot is generated
            for each column.
        dt_column: The datetime column used to analyze
            the temporal distribution.
        period: The temporal period e.g. monthly or daily.
        ncols: The number of columns.
        figsize: The figure size in inches. The first dimension
            is the width and the second is the height.

    Returns:
        The HTML representation of the figure.
    """
    if frame.shape[0] == 0:
        return ""
    figures = create_temporal_null_figures(
        frame=frame, columns=columns, dt_column=dt_column, period=period, figsize=figsize
    )
    figures = add_column_to_figure(columns=columns, figures=figures)
    return Template(
        """
    <div class="container-fluid text-center">
      <div class="row align-items-start">
        {{columns}}
      </div>
    </div>
    """
    ).render({"columns": "\n".join(split_figures_by_column(figures=figures, ncols=ncols))})


def create_temporal_null_figures(
    frame: pd.DataFrame,
    columns: Sequence[str],
    dt_column: str,
    period: str,
    figsize: tuple[float, float] = (7, 5),
) -> list[str]:
    r"""Create a HTML representation of each figure with the temporal
    null value distribution.

    Args:
        frame: The DataFrame to analyze.
        columns: The list of columns to analyze. A plot is generated
            for each column.
        dt_column: The datetime column used to analyze
            the temporal distribution.
        period: The temporal period e.g. monthly or daily.
        figsize: The figure size in inches. The first dimension
            is the width and the second is the height.

    Returns:
        The HTML representations of the figures.
    """
    if frame.shape[0] == 0:
        return []

    figures = []
    for column in tqdm(columns, desc="generating figures"):
        fig, ax = plt.subplots(figsize=figsize)
        ax.set_title(f"column: {column}")

        num_nulls, total, labels = prepare_data(
            frame=frame, column=column, dt_column=dt_column, period=period
        )
        plot_temporal_null_total(ax=ax, labels=labels, num_nulls=num_nulls, total=total)
        readable_xticklabels(ax, max_num_xticks=50)
        figures.append(figure2html(fig, close_fig=True))

    return figures


def add_column_to_figure(columns: Sequence[str], figures: Sequence[str]) -> list[str]:
    r"""Add the column name to the HTML representation of the figure.

    Args:
        columns: The column names.
        figures: The HTML representations of each figure.

    Returns:
        The updated HTML representations of each figure.

    Raises:
        RuntimeError: if the number of column names is different from
            the number of figures.
    """
    if len(columns) != len(figures):
        msg = (
            f"The number of column names is different from the number of figures: "
            f"{len(columns):,} vs{len(figures):,}"
        )
        raise RuntimeError(msg)
    outputs = []
    for i, (col, figure) in enumerate(zip(columns, figures)):
        outputs.append(f'<div style="text-align:center">({i}) {col}\n{figure}</div>')
    return outputs


def split_figures_by_column(figures: Sequence[str], ncols: int) -> list[str]:
    r"""Split the figures into multiple columns.

    Args:
        figures: The HTML representations of each figure.
        ncols: The number of columns.

    Returns:
        The columns.
    """
    cols = []
    for i in range(ncols):
        figs = str_indent("\n<hr>\n".join(figures[i::ncols]))
        cols.append(f'<div class="col">\n  {figs}\n</div>')
    return cols


def prepare_data(
    frame: pd.DataFrame,
    column: str,
    dt_column: str,
    period: str,
) -> tuple[np.ndarray, np.ndarray, list]:
    r"""Prepare the data to create the figure and table.

    Args:
        frame: The DataFrame to analyze.
        column: The column to analyze.
        dt_column: The datetime column used to analyze
            the temporal distribution.
        period: The temporal period e.g. monthly or
            daily.

    Returns:
        A tuple with 3 values. The first value is a numpy NDArray
            that contains the number of null values per period. The
            second value is a numpy NDArray that contains the total
            number of values. The third value is a list that contains
            the label of each period.

    Example usage:

    ```pycon
    >>> import numpy as np
    >>> import pandas as pd
    >>> from flamme.section.null_temp_col import prepare_data
    >>> num_nulls, total, labels = prepare_data(
    ...     frame=pd.DataFrame(
    ...         {
    ...             "col": np.array([np.nan, 1, 0, 1]),
    ...             "datetime": pd.to_datetime(
    ...                 ["2020-01-03", "2020-02-03", "2020-03-03", "2020-04-03"]
    ...             ),
    ...         }
    ...     ),
    ...     column="col",
    ...     dt_column="datetime",
    ...     period="M",
    ... )
    >>> num_nulls
    array([1, 0, 0, 0])
    >>> total
    array([1, 1, 1, 1])
    >>> labels
    ['2020-01', '2020-02', '2020-03', '2020-04']

    ```
    """
    dataframe = frame[[column, dt_column]].copy()
    dt_col = "__datetime__"
    dataframe[dt_col] = dataframe[dt_column].dt.to_period(period)

    null_col = f"__{column}_isna__"
    dataframe.loc[:, null_col] = dataframe.loc[:, column].isna()

    num_nulls = dataframe.groupby(dt_col)[null_col].sum().sort_index()
    total = dataframe.groupby(dt_col)[null_col].count().sort_index()
    labels = [str(dt) for dt in num_nulls.index]
    return num_nulls.to_numpy().astype(int), total.to_numpy().astype(int), labels


def create_table_section(
    frame: pd.DataFrame, columns: Sequence[str], dt_column: str, period: str
) -> str:
    r"""Create a HTML representation of a table with the temporal
    distribution of null values.

    Args:
        frame: The DataFrame to analyze.
        columns: The list of columns to analyze.
        dt_column: The datetime column used to analyze
            the temporal distribution.
        period: The temporal period e.g. monthly or
            daily.

    Returns:
        The HTML representation of the table.
    """
    if frame.shape[0] == 0:
        return ""
    tables = []
    for column in columns:
        table = create_temporal_null_table(
            frame=frame, column=column, dt_column=dt_column, period=period
        )
        tables.append(f'<p style="margin-top: 1rem;">\n\n{table}\n')
    return Template(
        """
<details>
    <summary>[show statistics per temporal period]</summary>

    <p style="margin-top: 1rem;">
    The following table shows some statistics for each period of column {{column}}.

    {{tables}}
</details>
"""
    ).render({"tables": "\n".join(tables)})


def create_temporal_null_table(
    frame: pd.DataFrame, column: str, dt_column: str, period: str
) -> str:
    r"""Create a HTML representation of a table with the temporal
    distribution of null values.

    Args:
        frame: The DataFrame to analyze.
        column: The column to analyze.
        dt_column: The datetime column used to analyze
            the temporal distribution.
        period: The temporal period e.g. monthly or
            daily.

    Returns:
        The HTML representation of the table.
    """
    if frame.shape[0] == 0:
        return ""
    num_nulls, totals, labels = prepare_data(
        frame=frame, column=column, dt_column=dt_column, period=period
    )
    rows = []
    for label, num_null, total in zip(labels, num_nulls, totals):
        rows.append(create_temporal_null_table_row(label=label, num_nulls=num_null, total=total))
    return Template(
        """
<table class="table table-hover table-responsive w-auto" >
    <thead class="thead table-group-divider">
        <tr>
            <th colspan="6" style="text-align: center">column: {{column}}</th>
        </tr>
        <tr>
            <th>period</th>
            <th>number of null values</th>
            <th>number of non-null values</th>
            <th>total number of values</th>
            <th>percentage of null values</th>
            <th>percentage of non-null values</th>
        </tr>
    </thead>
    <tbody class="tbody table-group-divider">
        {{rows}}
        <tr class="table-group-divider"></tr>
    </tbody>
</table>
"""
    ).render({"rows": "\n".join(rows), "column": column, "period": period})


def create_temporal_null_table_row(label: str, num_nulls: int, total: int) -> str:
    r"""Create the HTML code of a new table row.

    Args:
        label: The label of the row.
        num_nulls: The number of null values.
        total: The total number of values.

    Returns:
        The HTML code of a row.
    """
    num_non_nulls = total - num_nulls
    return Template(
        """<tr>
    <th>{{label}}</th>
    <td {{num_style}}>{{num_nulls}}</td>
    <td {{num_style}}>{{num_non_nulls}}</td>
    <td {{num_style}}>{{total}}</td>
    <td {{num_style}}>{{num_nulls_pct}}</td>
    <td {{num_style}}>{{num_non_nulls_pct}}</td>
</tr>"""
    ).render(
        {
            "num_style": 'style="text-align: right;"',
            "label": label,
            "num_nulls": f"{num_nulls:,}",
            "num_non_nulls": f"{num_non_nulls:,}",
            "total": f"{total:,}",
            "num_nulls_pct": f"{100 * num_nulls / total:.2f}%",
            "num_non_nulls_pct": f"{100 * num_non_nulls / total:.2f}%",
        }
    )
